"""
Load lattices from Matlab files.
"""

from __future__ import print_function, annotations

__all__ = ["load_mat", "save_mat", "load_m", "save_m", "load_var"]

import sys
import os
from os.path import abspath, basename, splitext
from typing import Optional, Any
from collections.abc import Sequence, Generator
from warnings import warn

import numpy as np
import scipy.io

# imports necessary in 'globals()' for 'eval'
from numpy import array, uint8, nan as NaN  # noqa: F401

from .allfiles import register_format
from .utils import split_ignoring_parentheses, RingParam, keep_elements
from .utils import _drop_attrs, _CLASS_MAP
from ..lattice import Element, Lattice, Particle, Filter
from ..lattice import elements, AtWarning, params_filter, AtError

# Translation of RingParam attributes
_m2p = {
    "FamName": "name",
    "Energy": "energy",
    "Periodicity": "periodicity",
    "Particle": "particle",
    "cell_harmnumber": "cell_harmnumber",  # necessary: property
    "beam_current": "beam_current",  # necessary: property
    "PassMethod": None,
    "Length": None,
    "cavpts": None,
}
_p2m = dict((v, k) for k, v in _m2p.items() if v is not None)
# Attribute to drop when writing a file
_p2m.update(_drop_attrs)

# Python to Matlab type translation
_mattype_map = {
    int: float,
    np.ndarray: lambda attr: np.asanyarray(attr),
    Particle: lambda attr: attr.to_dict(),
}
# Matlab constructor function
# Default: "".join(("at", element_class.__name__.lower()))
_mat_constructor = {
    "Dipole": "atsbend",
    "M66": "atM66",
}


def _mat_encoder(v):
    """type encoding for .mat files"""
    return _mattype_map.get(type(v), lambda attr: attr)(v)


def _matfile_generator(
    params: dict[str, Any], mat_file: str
) -> Generator[Element, None, None]:
    """Run through Matlab cells and generate AT elements"""

    def mclean(data):
        if data.dtype.type is np.str_:
            # Convert strings in arrays back to strings.
            return str(data[0]) if data.size > 0 else ""
        elif data.size == 1:
            v = data[0, 0]
            if issubclass(v.dtype.type, np.void):
                # Object => Return a dict
                return {f: mclean(v[f]) for f in v.dtype.fields}
            else:
                # Return a scalar
                return v
        else:
            # Remove any surplus dimensions in arrays.
            return np.squeeze(data)

    # noinspection PyUnresolvedReferences
    m = scipy.io.loadmat(params.setdefault("in_file", mat_file))
    matvars = [varname for varname in m if not varname.startswith("__")]
    default_key = matvars[0] if (len(matvars) == 1) else "RING"
    key = params.setdefault("use", default_key)
    if key not in m.keys():
        kok = [k for k in m.keys() if "__" not in k]
        raise AtError(
            f"Selected '{key}' variable does not exist, please select in: {kok}"
        )
    check = params.pop("check", True)
    quiet = params.pop("quiet", False)
    cell_array = m[key].flat
    for index, mat_elem in enumerate(cell_array):
        elem = mat_elem[0, 0]
        kwargs = {f: mclean(elem[f]) for f in elem.dtype.fields}
        yield Element.from_dict(kwargs, index=index, check=check, quiet=quiet)


def ringparam_filter(
    params: dict[str, Any], elem_iterator: Filter, *args
) -> Generator[Element, None, None]:
    """Run through all elements, process and optionally removes
    RingParam elements

    Parameters:
        params:         Lattice building parameters (see :py:class:`.Lattice`)
        elem_iterator:  Iterator over the lattice Elements

    Yields:
        elem (Element): new Elements

    The following keys in ``params`` are used:

    ============    ===================
    **keep_all**    keep RingParam elem_iterator as Markers
    ============    ===================

    The following keys in ``params`` are set:

    * ``name``
    * ``energy``
    * ``periodicity``
    * ``_harmnumber`` or
    * ``harmonic_number``
    * ``_particle``
    * ``_radiation``
    """
    keep_all = params.pop("keep_all", False)
    ringparams = []
    radiate = False
    for elem in elem_iterator(params, *args):
        if elem.PassMethod.endswith("RadPass") or elem.PassMethod.endswith(
            "CavityPass"
        ):
            radiate = True
        if isinstance(elem, RingParam):
            ringparams.append(elem)
            for k, v in elem.items():
                k2 = _m2p.get(k, k)
                if k2 is not None:
                    params.setdefault(k2, v)
            if keep_all:
                pars = vars(elem).copy()
                name = pars.pop("FamName")
                yield elements.Marker(name, tag="RingParam", **pars)
        else:
            yield elem
    params["_radiation"] = radiate

    if len(ringparams) > 1:
        warn(AtWarning("More than 1 RingParam element, the 1st one is used"))


def load_mat(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a Matlab mat-file

    Parameters:
        filename:           Name of a '.mat' file

    Keyword Args:
        use (str):          Name of the Matlab variable containing
          the lattice. Default: it there is a single variable, use it, otherwise
          select 'RING'
        mat_key (str):      deprecated alias for *use*
        check (bool):       Run the coherence tests. Default:
          :py:obj:`True`
        quiet (bool):       Suppress the warning for non-standard
          classes. Default: :py:obj:`False`
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    if "key" in kwargs:  # process the deprecated 'key' keyword
        kwargs.setdefault("use", kwargs.pop("key"))
    if "mat_key" in kwargs:  # process the deprecated 'mat_key' keyword
        kwargs.setdefault("use", kwargs.pop("mat_key"))
    return Lattice(
        ringparam_filter,
        _matfile_generator,
        abspath(filename),
        iterator=params_filter,
        **kwargs,
    )


def _element_from_m(line: str) -> Element:
    """Builds an :py:class:`.Element` from a line in an m-file

    Parameters:
        line:           Matlab string representation of an :py:class:`.Element`

    Returns:
        elem (Element): new :py:class:`.Element`
    """

    def argsplit(value):
        return [a.strip() for a in split_ignoring_parentheses(value)]

    def makedir(mat_struct):
        """Build directory from Matlab struct arguments"""

        def pairs(it):
            while True:
                try:
                    a = next(it)
                except StopIteration:
                    break
                yield eval(a), convert(next(it))

        return dict(pairs(iter(mat_struct)))

    def makearray(mat_arr):
        """Build numpy array for Matlab array syntax"""

        def arraystr(arr):
            lns = arr.split(";")
            rr = [arraystr(v) for v in lns] if len(lns) > 1 else lns[0].split()
            return f"[{', '.join(rr)}]"

        return eval(f"array({arraystr(mat_arr)})")

    def convert(value):
        """convert Matlab syntax to numpy syntax"""
        if value.startswith("["):
            result = makearray(value[1:-1])
        elif value.startswith("struct"):
            result = makedir(argsplit(value[7:-1]))
        else:
            result = eval(value)
        return result

    left = line.index("(")
    right = line.rindex(")")
    matcls = line[:left].strip()[2:]
    cls = _CLASS_MAP[matcls]
    arguments = argsplit(line[left + 1 : right])
    ll = len(cls._BUILD_ATTRIBUTES)
    if ll < len(arguments) and arguments[ll].endswith("Pass'"):
        arguments.insert(ll, "'PassMethod'")
    args = [convert(v) for v in arguments[:ll]]
    kwargs = makedir(arguments[ll:])
    if matcls == "rbend":
        # the Matlab 'rbend' has no equivalent in PyAT. This adds parameters
        # necessary for using the python sector bend
        halfangle = 0.5 * args[2]
        kwargs.setdefault("EntranceAngle", halfangle)
        kwargs.setdefault("ExitAngle", halfangle)
    return cls(*args, **kwargs)


def load_m(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a Matlab m-file

    Parameters:
        filename:           Name of a '.m' file

    Keyword Args:
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """

    def mfile_generator(params: dict, m_file: str) -> Generator[Element, None, None]:
        """Run through the lines of a Matlab m-file and generate AT elements"""
        with open(params.setdefault("in_file", m_file), "rt") as file:
            _ = next(file)  # Matlab function definition
            _ = next(file)  # Cell array opening
            for lineno, line in enumerate(file):
                if line.startswith("};"):
                    break
                try:
                    elem = _element_from_m(line)
                except ValueError:
                    warn(AtWarning("Invalid line {0} skipped.".format(lineno)))
                    continue
                except KeyError:
                    warn(AtWarning("Line {0}: Unknown class.".format(lineno)))
                    continue
                else:
                    yield elem

    return Lattice(
        ringparam_filter,
        mfile_generator,
        abspath(filename),
        iterator=params_filter,
        **kwargs,
    )


def load_var(matlat: Sequence[dict], **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice` from a Matlab cell array

    Parameters:
        matlat:             Matlab lattice

    Keyword Args:
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object
    """

    # noinspection PyUnusedLocal
    def var_generator(params, latt):
        for elem in latt:
            yield Element.from_dict(elem)

    return Lattice(
        ringparam_filter, var_generator, matlat, iterator=params_filter, **kwargs
    )


def matlab_ring(ring: Lattice) -> Generator[Element, None, None]:
    """Prepend a RingParam element to a lattice"""

    def required(rng):
        # Public lattice attributes
        params = dict((k, v) for k, v in vars(rng).items() if not k.startswith("_"))
        # Output the required attributes/properties
        for kp, km in _p2m.items():
            try:
                v = getattr(rng, kp)
            except AttributeError:
                pass
            else:
                params.pop(kp, None)
                if km is not None:
                    yield km, v
        # Output the remaining attributes
        yield from params.items()

    dct = dict(required(ring))
    yield RingParam(**dct)
    yield from keep_elements(ring)


def save_mat(ring: Lattice, filename: str, **kwargs) -> None:
    """Save a :py:class:`.Lattice` as a Matlab mat-file

    Parameters:
        ring:       Lattice description
        filename:   Name of the '.mat' file

    Keyword Args:
        use (str):  Name of the Matlab variable containing the lattice, Default: "RING"
        mat_key (str): Deprecated, alias for *use*

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """
    # Ensure the lattice is a Matlab column vector: list(list)
    use = kwargs.pop("mat_key", "RING")  # For backward compatibility
    use = kwargs.pop("use", use)
    lring = [[el.to_dict(encoder=_mat_encoder)] for el in matlab_ring(ring)]
    scipy.io.savemat(filename, {use: lring}, long_field_names=True)


def _element_to_m(elem: Element) -> str:
    """Builds the Matlab-evaluable string for an :py:class:`.Element`

    Parameters:
        elem:           :py:class:`.Element`

    Returns:
        mstr (str):     Matlab string representation of the
          :py:class:`.Element` attributes
    """

    def convert(arg):
        def convert_dict(pdir):
            def scan(d):
                for k, v in d.items():
                    yield convert(k)
                    yield convert(v)

            return "struct({0})".format(", ".join(scan(pdir)))

        def convert_array(arr):
            if arr.ndim > 1:
                lns = (str(list(ln)).replace(",", "")[1:-1] for ln in arr)
                return "".join(("[", "; ".join(lns), "]"))
            elif arr.ndim > 0:
                return str(list(arr)).replace(",", "")
            else:
                return str(arr)

        if isinstance(arg, np.ndarray):
            return convert_array(arg)
        elif isinstance(arg, dict):
            return convert_dict(arg)
        elif isinstance(arg, Particle):
            return convert_dict(arg.to_dict())
        else:
            return repr(arg)

    def m_name(elclass):
        classname = elclass.__name__
        return _mat_constructor.get(classname, "".join(("at", classname.lower())))

    attrs = dict(elem.items())
    # noinspection PyProtectedMember
    args = [attrs.pop(k, getattr(elem, k)) for k in elem._BUILD_ATTRIBUTES]
    defelem = elem.__class__(*args)
    kwds = dict(
        (k, v)
        for k, v in attrs.items()
        if not np.array_equal(v, getattr(defelem, k, None))
    )
    argstrs = [convert(arg) for arg in args]
    if "PassMethod" in kwds:
        argstrs.append(convert(kwds.pop("PassMethod")))
    argstrs += [", ".join((repr(k), convert(v))) for k, v in kwds.items()]
    return "{0:>15}({1});...".format(m_name(elem.__class__), ", ".join(argstrs))


def save_m(ring: Lattice, filename: Optional[str] = None) -> None:
    """Save a :py:class:`.Lattice` as a Matlab m-file

    Parameters:
        ring:           Lattice description
        filename:       Name of the '.m' file. Default: outputs on
          :py:obj:`sys.stdout`

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """

    def save(file):
        print("ring = {...", file=file)
        for elem in matlab_ring(ring):
            print(_element_to_m(elem), file=file)
        print("};", file=file)

    if filename is None:
        save(sys.stdout)
    else:
        with open(filename, "wt") as mfile:
            [funcname, _] = splitext(basename(filename))
            print("function ring = {0}()".format(funcname), file=mfile)
            save(mfile)
            print("end", file=mfile)


# Simulates the deprecated "mat_file" and "mat_key" attributes
def _mat_file(ring):
    """.mat input file. Deprecated, use *in_file* instead."""
    try:
        in_file = ring.in_file
    except AttributeError:
        raise AttributeError("'Lattice' object has no attribute 'mat_file'")
    else:
        _, ext = os.path.splitext(in_file)
        if ext != ".mat":
            raise AttributeError("'Lattice' object has no attribute 'mat_file'")
    return in_file


def _mat_key(ring):
    """selected Matlab variable. Deprecated, use *use* instead."""
    try:
        mat_key = ring.use
    except AttributeError:
        raise AttributeError("'Lattice' object has no attribute 'mat_key'")
    return mat_key


register_format(
    ".mat",
    load_mat,
    save_mat,
    descr="Matlab binary mat-file. See :py:func:`.load_mat`.",
)

register_format(
    ".m",
    load_m,
    save_m,
    descr="Matlab text m-file. See :py:func:`.load_m`.",
)

Lattice.mat_file = property(_mat_file, None, None)
Lattice.mat_key = property(_mat_key, None, None)



from mpi4py import MPI
import numpy

def mpi_sweep(generator, worker, collector):
    """
    Evaluate function on multiple MPI nodes and collect the result.
    generator: function returning list of worker arguments
    worker: worker function
    collector: function collecting calculation results
    """
    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    rank = comm.Get_rank()
    name = MPI.Get_processor_name()

    data = None
    if rank == 0:
        data = numpy.array_split(generator(), size)

    chunk = comm.scatter(data)

    output = [worker(par, name=name, rank=rank, size=size)
              for par in chunk]

    data = comm.gather(list(zip(chunk, output)))

    if rank == 0:
        return collector([d for sublist in data for d in sublist])



"""Stub file for the 'atpass' extension"""

import numpy as np
from typing import List, Optional
from at.lattice import Element, Particle

def atpass(line: List[Element], r_in: np.ndarray, nturns: int,
           refpts: np.ndarray,
           turn: Optional[int] = None,
           energy: Optional[float] = None,
           particle: Optional[Particle] = None,
           keep_counter: bool = False,
           reuse: bool = False,
           omp_num_thread: int = 0,
           losses: bool = False,
           bunch_spos = None, bunch_current = None): ...

def elempass(element: Element, r_in,
             energy: Optional[float] = None,
             particle: Optional[Particle] = None,
             ): ...

def reset_rng(*, rank: int = 0, seed: Optional[int] = None) -> None: ...
def common_rng() -> float: ...
def thread_rng() -> float: ...



"""Stub file for the 'diffmatrix' extension"""

import numpy as np
from typing import Optional
from at.lattice import Element
from . import Orbit

def find_mpole_raddiff_matrix(element: Element, orbit: Orbit,
                              energy: Optional[float] = None) -> np.ndarray: ...


from __future__ import annotations
import numpy
from .atpass import atpass as _atpass, elempass as _elempass
from .utils import fortran_align, has_collective, format_results
from .utils import initialize_lpass, disable_varelem, variable_refs
from ..lattice import Lattice, Element, Refpts, End
from ..lattice import get_uint32_index
from ..lattice import AtWarning, DConstant, random
from collections.abc import Iterable
from typing import Optional
from functools import partial
import multiprocessing
from warnings import warn
from .atpass import reset_rng


__all__ = ['lattice_track', 'element_track', 'internal_lpass',
           'internal_epass', 'internal_plpass']

_imax = numpy.iinfo(int).max
_globring: Optional[list[Element]] = None


def _atpass_fork(seed, rank, rin, **kwargs):
    """Single forked job"""
    reset_rng(rank=rank, seed=seed)
    result = _atpass(_globring, rin, **kwargs)
    return rin, result


def _atpass_spawn(ring, seed, rank, rin, **kwargs):
    """Single spawned job"""
    reset_rng(rank=rank, seed=seed)
    result = _atpass(ring, rin, **kwargs)
    return rin, result


def _pass(ring, r_in, pool_size, start_method, **kwargs):
    ctx = multiprocessing.get_context(start_method)
    # Split input in as many slices as processes
    args = enumerate(numpy.array_split(r_in, pool_size, axis=1))
    # Generate a new starting point for C RNGs
    seed = random.common.integers(0, high=_imax, dtype=int)
    global _globring
    _globring = ring
    if ctx.get_start_method() == 'fork':
        passfunc = partial(_atpass_fork, seed, **kwargs)
    else:
        passfunc = partial(_atpass_spawn, ring, seed, **kwargs)
    # Start the parallel jobs
    with ctx.Pool(pool_size) as pool:
        results = pool.starmap(passfunc, args)
    _globring = None
    # Gather the results
    losses = kwargs.pop('losses', False)
    return format_results(results, r_in, losses)


@fortran_align
def _element_pass(element: Element, r_in, **kwargs):
    return _elempass(element, r_in, **kwargs)


@fortran_align
def _lattice_pass(lattice: list[Element], r_in, nturns: int = 1,
                  refpts: Refpts = End, no_varelem=True, **kwargs):
    kwargs['reuse'] = kwargs.pop('keep_lattice', False)
    if no_varelem:
        lattice = disable_varelem(lattice)
    else:
        if sum(variable_refs(lattice)) > 0:
            kwargs['reuse'] = False
    refs = get_uint32_index(lattice, refpts)
    return _atpass(lattice, r_in, nturns, refpts=refs, **kwargs)


@fortran_align
def _plattice_pass(lattice: list[Element], r_in, nturns: int = 1,
                   refpts: Refpts = End, pool_size: int = None,
                   start_method: str = None, **kwargs):
    refpts = get_uint32_index(lattice, refpts)
    any_collective = has_collective(lattice)
    kwargs['reuse'] = kwargs.pop('keep_lattice', False)
    rshape = r_in.shape
    if len(rshape) >= 2 and rshape[1] > 1 and not any_collective:
        if pool_size is None:
            pool_size = min(len(r_in[0]), multiprocessing.cpu_count(),
                            DConstant.patpass_poolsize)
        if start_method is None:
            start_method = DConstant.patpass_startmethod
        return _pass(lattice, r_in, pool_size, start_method, nturns=nturns,
                     refpts=refpts, **kwargs)
    else:
        if any_collective:
            warn(AtWarning('Collective PassMethod found: use single process'))
        else:
            warn(AtWarning('no parallel computation for a single particle'))
        return _atpass(lattice, r_in, nturns=nturns, refpts=refpts, **kwargs)


def lattice_track(lattice: Iterable[Element], r_in,
                  nturns: int = 1, refpts: Refpts = End,
                  in_place: bool = False, **kwargs):
    """
    :py:func:`track_function` tracks particles through each element of a
    lattice or throught a single Element calling the element-specific
    tracking function specified in the Element's *PassMethod* field.

    Usage:
      >>> lattice_track(lattice, r_in)
      >>> lattice.track(r_in)

    Parameters:
        lattice: list of elements
        r_in: (6, N) array: input coordinates of N particles.
          *r_in* is modified in-place only if *in_place* is 
          :py:obj:`True` and reports the coordinates at
          the end of the element. For the best efficiency, *r_in*
          should be given as F_CONTIGUOUS numpy array.

    Keyword arguments:
        nturns: number of turns to be tracked
        refpts: Selects the location of coordinates output.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        in_place (bool): If True *r_in* is modified in-place and
          reports the coordinates at the end of the element.
          (default: False)
        keep_lattice (bool):    Use elements persisted from a previous
          call. If :py:obj:`True`, assume that the lattice has not changed
          since the previous call.
        keep_counter (bool):    Keep the turn number from the previous
          call.
        turn (int):             Starting turn number. Ignored if
          *keep_counter* is :py:obj:`True`. The turn number is necessary to
          compute the absolute path length used in RFCavityPass.
        losses (bool):          Boolean to activate loss maps output
        omp_num_threads (int):  Number of OpenMP threads
          (default: automatic)
        use_mp (bool): Flag to activate multiprocessing (default: False)
        pool_size:              number of processes used when
          *use_mp* is :py:obj:`True`. If None, ``min(npart,nproc)``
          is used. It can be globally set using the variable
          *at.lattice.DConstant.patpass_poolsize*
        start_method:           python multiprocessing start method.
          :py:obj:`None` uses the python default that is considered safe.
          Available values: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          Default for linux is ``'fork'``, default for macOS and  Windows is
          ``'spawn'``. ``'fork'`` may be used on macOS to speed up the
          calculation or to solve Runtime Errors, however it is considered
          unsafe. Used only when *use_mp* is :py:obj:`True`. It can be globally
          set using the variable *at.lattice.DConstant.patpass_startmethod*

    The following keyword arguments overload the lattice values

    Keyword arguments:

        particle (Optional[Particle]): circulating particle.
          Default: :code:`lattice.particle` if existing,
          otherwise :code:`Particle('relativistic')`
        energy (Optiona[float]): lattice energy. Default 0.
        unfold_beam (bool): Internal beam folding activate, this
          assumes the input particles are in bucket 0, works only
          if all bucket see the same RF Voltage.
          Default: :py:obj:`True`

    If *energy* is not available, relativistic tracking if forced,
    *rest_energy* is ignored.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns
        trackparam: A dictionary containing tracking input parameters with the
          following keys:

          ==============    ===================================================
          **npart**         number of particles
          **rout**          final particle coordinates
          **turn**          starting turn
          **refpts**        array of index where particle coordinate are saved
                            (only for lattice tracking)
          **nturns**        number of turn
          ==============    ===================================================

        trackdata: A dictionary containing tracking data with the following
          keys:

          ==============    ===================================================
          **loss_map**:     recarray containing the loss_map (only for lattice
                            tracking)
          ==============    ===================================================


          The **loss_map** is filled only if *losses* is :py:obj:`True`,
          it contains the following keys:

          ==============    ===================================================
          **islost**        (npart,) bool array indicating lost particles
          **turn**          (npart,) int array indicating the turn at
                            which the particle is lost
          **element**       (npart,) int array indicating the element at
                            which the particle is lost
          **coord**         (npart, 6) float array giving the coordinates at
                            which the particle is lost (zero for surviving
                            particles)
          ==============    ===================================================


    .. note::

       * :pycode:`track_function(lattice, r_in, refpts=len(line))` is the same
         as :pycode:`track_function(lattice, r_in)` since the reference point
         len(line) is the exit of the last element.
       * :pycode:`track_function(lattice, r_in, refpts=0)` is a copy of *r_in*
         since the reference point 0 is the entrance of the first element.
       * To resume an interrupted tracking (for instance to get intermediate
         results), one must use one of the *turn* or *keep_counter*
         keywords to ensure the continuity of the turn number.
       * For multiparticle tracking with large number of turn the size of
         *r_out* may increase excessively. To avoid memory issues
         :pycode:`track_function(lattice, r_in, refpts=None, in_place=True)`
         can be used. An empty list is returned and the tracking results of
         the last turn are stored in *r_in*.
       * To model buckets with different RF voltage :pycode:`unfold_beam=False`
         has to be used. The beam can be unfolded using the function
         :py:func:`.unfold_beam`. This function takes into account
         the true voltage in each bucket and distributes the particles in the
         bunches defined by :code:`ring.fillpattern` using a 6D orbit search.
    """
    trackdata = {}
    trackparam = {}
    part_kw = ['energy', 'particle']
    try:
        npart = numpy.shape(r_in)[1]
    except IndexError:
        npart = 1

    [trackparam.update((kw, kwargs.get(kw))) for kw in kwargs if kw in part_kw]
    trackparam.update({'npart': npart})

    if not in_place:
        r_in = r_in.copy()

    lattice = initialize_lpass(lattice, nturns, kwargs)
    ldtype = [('islost', numpy.bool_),
              ('turn', numpy.uint32),
              ('elem', numpy.uint32),
              ('coord', numpy.float64, (6,)),
              ]
    loss_map = numpy.recarray((npart,), ldtype)
    lat_kw = ['turn']
    [trackparam.update((kw, kwargs.get(kw)))
     for kw in kwargs if kw in lat_kw]
    trackparam.update({'refpts': get_uint32_index(lattice, refpts),
                       'nturns': nturns})

    use_mp = kwargs.pop('use_mp', False)
    start_method = kwargs.pop('start_method', None)
    pool_size = kwargs.pop('pool_size', None)
    if use_mp:
        kwargs.update({'pool_size': pool_size,
                       'start_method': start_method})
        rout = _plattice_pass(lattice, r_in, nturns=nturns,
                              refpts=refpts, **kwargs)
    else:
        rout = _lattice_pass(lattice, r_in, nturns=nturns,
                             refpts=refpts, no_varelem=False,
                             **kwargs)

    if kwargs.get('losses', False):
        rout, lm = rout
        lm['coord'] = lm['coord'].T
        for k, v in lm.items():
            loss_map[k] = v

    trackdata.update({'loss_map': loss_map})
    trackparam.update({'rout': r_in})

    return rout, trackparam, trackdata


def element_track(element: Element, r_in, in_place: bool = False, **kwargs):
    """
    :py:func:`element_track` tracks particles through one element of a
    calling the element-specific tracking function specified in the
    Element's *PassMethod* field

    Usage:
      >>> element_track(element, r_in)
      >>> element.track(r_in)

    Parameters:
        element: element to track through
        r_in: (6, N) array: input coordinates of N particles.
          For the best efficiency, *r_in*
          should be given as F_CONTIGUOUS numpy array.

    Keyword arguments:
        in_place (bool): If True *r_in* is modified in-place and
          reports the coordinates at the end of the element.
          (default: False)
        omp_num_threads (int):  Number of OpenMP threads
          (default: automatic)
        particle (Optional[Particle]): circulating particle.
          Default: :code:`lattice.particle` if existing,
          otherwise :code:`Particle('relativistic')`
        energy (Optiona[float]): lattice energy. Default 0.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns
    """
    if not in_place:
        r_in = r_in.copy()

    rout = _element_pass(element, r_in, **kwargs)
    return rout
    

internal_lpass = _lattice_pass
internal_epass = _element_pass
internal_plpass = _plattice_pass
Lattice.track = lattice_track
Element.track = element_track



r"""
Helper functions for working with AT lattices.

A lattice as understood by pyAT is a sequence of elements.  These functions
are useful for working with these sequences.

.. _refpts:

**Selecting elements in a lattice:**

The *refpts* argument allows functions to select locations in the lattice. The
location is defined as the entrance of the selected element. *refpts* may be:

#. an integer in the range *[-len(lattice), len(lattice)-1]*
   selecting an element according to the python indexing rules.
   As a special case, *len(lattice)* is allowed and refers to the end
   of the last element,
#. an ordered list of such integers without duplicates,
#. a numpy array of booleans of maximum length *len(lattice)+1*,
   where selected elements are :py:obj:`True`,
#. :py:obj:`None`, meaning an empty selection,
#. :py:obj:`.All`, meaning "all possible reference points": the entrance of all
   elements plus the end of the last element,
#. :py:obj:`.End`, selecting the end of the last element,
#. an element type, selecting all the elements of that type in
   the lattice, e.g. :pycode:`at.Sextupole`,
#. a string, selecting all the elements whose `FamName` attribute matches it.
   Unix shell-style wildcards are accepted, e.g. `"Q[FD]*"`,
#. a callable :pycode:`filtfunc` such that :pycode:`filtfunc(elem)`
   is :py:obj:`True` for selected elements.

"""
import numpy
import functools
import re
from typing import Callable, Optional, Sequence, Iterator
from typing import Union, Tuple, List, Type
from enum import Enum
from itertools import compress
from operator import attrgetter
from fnmatch import fnmatch
from .elements import Element, Dipole

_GEOMETRY_EPSIL = 1.0e-3

ElementFilter = Callable[[Element], bool]
BoolRefpts = numpy.ndarray
Uint32Refpts = numpy.ndarray


__all__ = ['All', 'End', 'AtError', 'AtWarning', 'axis_descr',
           'check_radiation', 'check_6d',
           'set_radiation', 'set_6d',
           'make_copy', 'uint32_refpts', 'bool_refpts',
           'get_uint32_index', 'get_bool_index',
           'checkattr', 'checktype', 'checkname',
           'get_elements', 'get_s_pos',
           'refpts_count', 'refpts_iterator',
           'set_shift', 'set_tilt', 'set_rotation',
           'tilt_elem', 'shift_elem', 'rotate_elem',
           'get_value_refpts', 'set_value_refpts', 'Refpts',
           'get_geometry', 'setval', 'getval']

_axis_def = dict(
    x=dict(index=0, label="x", unit=" [m]"),
    xp=dict(index=1, label="x'", unit=" [rad]"),
    y=dict(index=2, label="y", unit=" [m]"),
    yp=dict(index=3, label="y'", unit=" [rad]"),
    dp=dict(index=4, label=r"$\delta$", unit=""),
    ct=dict(index=5, label=r"$\beta c \tau$", unit=" [m]"),
)
for vvv in [vv for vv in _axis_def.values()]:
    _axis_def[vvv['index']] = vvv


class AtError(Exception):
    pass


class AtWarning(UserWarning):
    pass


_typ1 = "None, All, End, int, bool"

_typ2 = "None, All, End, int, bool, str, Type[Element], ElementFilter"


class RefptsCode(Enum):
    All = 'All'
    End = 'End'


RefIndex = Union[None, int, Sequence[int], bool, Sequence[bool], RefptsCode]
Refpts = Union[Type[Element], Element, ElementFilter, str, RefIndex]


#: :py:obj:`All` is a special value to be used as *refpts*. It means
#: "all possible reference points": the entrance of all elements plus the end
#: of the last element.
All = RefptsCode.All

#: :py:obj:`End` is a special value to be used as *refpts*. It refers to the
#: end of the last element.
End = RefptsCode.End


def _type_error(refpts, types):
    if isinstance(refpts, numpy.ndarray):
        tp = refpts.dtype.type
    else:
        tp = type(refpts)
    return TypeError(
        "Invalid refpts type {0}. Allowed types: {1}".format(tp, types))


# setval and getval return pickleable functions: no inner, nested function
# are allowed. So nested functions are replaced be module-level callable
# class instances
class _AttrItemGetter(object):
    __slots__ = ["attrname", "index"]

    def __init__(self, attrname: str, index: int):
        self.attrname = attrname
        self.index = index

    def __call__(self, elem):
        return getattr(elem, self.attrname)[self.index]


def getval(attrname: str, index: Optional[int] = None) -> Callable:
    """Return a callable object which fetches item *index* of
    attribute *attrname* of its operand. Examples:

    - After ``f = getval('Length')``, ``f(elem)`` returns ``elem.Length``
    - After ``f = getval('PolynomB, index=1)``, ``f(elem)`` returns
      ``elem.PolynomB[1]``

    """
    if index is None:
        return attrgetter(attrname)
    else:
        return _AttrItemGetter(attrname, index)


class _AttrSetter(object):
    __slots__ = ["attrname"]

    def __init__(self, attrname: str):
        self.attrname = attrname

    def __call__(self, elem, value):
        setattr(elem, self.attrname, value)


class _AttrItemSetter(object):
    __slots__ = ["attrname", "index"]

    def __init__(self, attrname: str, index: int):
        self.attrname = attrname
        self.index = index

    def __call__(self, elem, value):
        getattr(elem, self.attrname)[self.index] = value


def setval(attrname: str, index: Optional[int] = None) -> Callable:
    """Return a callable object which sets the value of  item *index* of
    attribute *attrname* of its 1st argument to it 2nd orgument.

    - After ``f = setval('Length')``, ``f(elem, value)`` is equivalent to
      ``elem.Length = value``
    - After ``f = setval('PolynomB, index=1)``, ``f(elem, value)`` is
      equivalent to ``elem.PolynomB[1] = value``

    """
    if index is None:
        return _AttrSetter(attrname)
    else:
        return _AttrItemSetter(attrname, index)


# noinspection PyIncorrectDocstring
def axis_descr(*args, key=None) -> Tuple:
    r"""axis_descr(axis [ ,axis], key=None)

    Return a tuple containing for each input argument the requested information

    Parameters:
        axis (Union[int, str]):    either an index in 0:6 or a string in
          ['x', 'xp', 'y', 'yp', 'dp', 'ct']
        key:                        key in the coordinate description
          dictionary, selecting the desired information. One of :

          'index'
            index in the standard AT coordinate vector
          'label'
            label for plot annotation
          'unit'
            coordinate unit
          :py:obj:`None`
            entire description dictionary

    Returns:
        descr (Tuple): requested information for each input argument.

    Examples:

        >>> axis_descr('x','dp', key='index')
        (0, 4)

        returns the indices in the standard coordinate vector

        >>> dplabel, = axis_descr('dp', key='label')
        >>> print(dplabel)
        $\delta$

        returns the coordinate label for plot annotation

        >>> axis_descr('x','dp')
        ({'index': 0, 'label': 'x', 'unit': ' [m]'},
         {'index': 4, 'label': '$\\delta$', 'unit': ''})

        returns the entire description directories

    """
    if key is None:
        return tuple(_axis_def[k] for k in args)
    else:
        return tuple(_axis_def[k][key] for k in args)


def check_radiation(rad: bool) -> Callable:
    r"""Deprecated decorator for optics functions (see :py:func:`check_6d`).

    :meta private:
    """
    return check_6d(rad)


def set_radiation(rad: bool) -> Callable:
    r"""Deprecated decorator for optics functions (see :py:func:`set_6d`).

    :meta private:
    """
    return set_6d(rad)


def check_6d(is_6d: bool) -> Callable:
    r"""Decorator for optics functions

    Wraps a function like :pycode:`func(ring, *args, **kwargs)` where
    :pycode:`ring` is a :py:class:`.Lattice` object. Raise :py:class:`.AtError`
    if :pycode:`ring.is_6d` is not the desired *is_6d* state. No test is
    performed if :pycode:`ring` is not a :py:class:`.Lattice`.

    Parameters:
        is_6d: Desired 6D state

    Raises:
        AtError: if :pycode:`ring.is_6d` is not *is_6d*

    Example:

        >>> @check_6d(False)
        ... def find_orbit4(ring, dct=None, guess=None, **kwargs):
                ...

        Raises :py:class:`.AtError` if :pycode:`ring.is_6d` is :py:obj:`True`

    See Also:
        :py:func:`set_6d`
    """
    def radiation_decorator(func):
        @functools.wraps(func)
        def wrapper(ring, *args, **kwargs):
            ringrad = getattr(ring, 'is_6d', is_6d)
            if ringrad != is_6d:
                raise AtError('{0} needs "ring.is_6d" {1}'.format(
                    func.__name__, is_6d))
            return func(ring, *args, **kwargs)
        return wrapper
    return radiation_decorator


def set_6d(is_6d: bool) -> Callable:
    # noinspection PyUnresolvedReferences
    r"""Decorator for optics functions

    Wraps a function like :pycode:`func(ring, *args, **kwargs)` where
    :pycode:`ring` is a :py:class:`.Lattice` object. If :pycode:`ring.is_6d`
    is not the desired *is_6d* state, :pycode:`func()` will be called with a
    modified copy of :pycode:`ring` satisfying the *is_6d* state.

    Parameters:
        is_6d: Desired 6D state

    Example:

        >>> @set_6d(True)
        ... def compute(ring)
                ...

        is roughly equivalent to:

        >>> if ring.is_6d:
        ...     compute(ring)
        ... else:
        ...     compute(ring.enable_6d(copy=True))

        See Also:
            :py:func:`check_6d`, :py:meth:`.Lattice.enable_6d`,
            :py:meth:`.Lattice.disable_6d`
        """
    if is_6d:
        def setrad_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, *args, **kwargs):
                rg = ring if ring.is_6d else ring.enable_6d(copy=True)
                return func(rg, *args, **kwargs)
            return wrapper
    else:
        def setrad_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, *args, **kwargs):
                rg = ring.disable_6d(copy=True) if ring.is_6d else ring
                return func(rg, *args, **kwargs)
            return wrapper
    return setrad_decorator


def make_copy(copy: bool) -> Callable:
    r"""Decorator for optics functions

    Wraps a function like :pycode:`func(ring, refpts, *args, **kwargs)` where
    :pycode:`ring` is a :py:class:`.Lattice` object.

    If *copy* is :py:obj:`False`, the function is not modified,

    If *copy* is :py:obj:`True`, a shallow copy of :pycode:`ring` is done,
    then the elements selected by *refpts* are deep-copied, then :pycode:`func`
    is applied to the copy, and the new :pycode:`ring` is returned.

    Parameters:
        copy:   If :py:obj:`True`, apply the decorated function to a copy of
          :pycode:`ring`
    """
    if copy:
        def copy_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, refpts, *args, **kwargs):
                try:
                    ring = ring.replace(refpts)
                except AttributeError:
                    check = get_bool_index(ring, refpts)
                    ring = [el.deepcopy() if ok else el 
                            for el, ok in zip(ring, check)]
                func(ring, refpts, *args, **kwargs)
                return ring
            return wrapper
    else:
        def copy_decorator(func):
            return func
    return copy_decorator


def uint32_refpts(refpts: RefIndex, n_elements: int,
                  endpoint: bool = True,
                  types: str = _typ1) -> Uint32Refpts:
    r"""Return a :py:obj:`~numpy.uint32` array of element indices selecting
    ring elements.

    Parameters:
        refpts:     Element selector. *refpts* may be:

          #. an integer or a sequence of integers
             (0 indicating the first element),
          #. a sequence of booleans marking the selected elements,
          #. :py:obj:`None`, meaning empty selection,
          #. :py:obj:`.All`, meaning "all possible reference points",
          #. :py:obj:`.End`, selecting the end of the last element.
        n_elements: Length of the sequence of elements
        endpoint:   if :py:obj:`True`, allow *n_elements* as a
          special index, referring to the end of the last element.
        types:      Allowed types

    Returns:
        uint32_ref (Uint32Refpts):  :py:obj:`~numpy.uint32` numpy array used
          for indexing :py:class:`.Element`\ s in a lattice.
    """
    refs = numpy.ravel(refpts)
    if refpts is RefptsCode.All:
        stop = n_elements+1 if endpoint else n_elements
        return numpy.arange(stop, dtype=numpy.uint32)
    elif refpts is RefptsCode.End:
        if not endpoint:
            raise IndexError('"End" index out of range')
        return numpy.array([n_elements], dtype=numpy.uint32)
    elif (refpts is None) or (refs.size == 0):
        return numpy.array([], dtype=numpy.uint32)
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        return numpy.flatnonzero(refs).astype(numpy.uint32)
    elif numpy.issubdtype(refs.dtype, numpy.integer):

        # Handle negative indices
        if endpoint:
            refs = numpy.array([i if (i == n_elements) else i % n_elements
                                for i in refs], dtype=numpy.uint32)
        else:
            refs = numpy.array([i % n_elements
                                for i in refs], dtype=numpy.uint32)
        # Check ascending
        if refs.size > 1:
            prev = refs[0]
            for nxt in refs[1:]:
                if nxt < prev:
                    raise IndexError('Index out of range or not in ascending'
                                     ' order')
                elif nxt == prev:
                    raise IndexError('Duplicated index')
                prev = nxt

        return refs
    else:
        raise _type_error(refpts, types)


# noinspection PyIncorrectDocstring
def get_uint32_index(ring: Sequence[Element], refpts: Refpts,
                     endpoint: bool = True,
                     regex: bool = False) -> Uint32Refpts:
    # noinspection PyUnresolvedReferences, PyShadowingNames
    r"""Returns an integer array of element indices, selecting ring elements.

    Parameters:
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        endpoint:   if :py:obj:`True`, allow *len(ring)* as a
          special index, referring to the end of the last element.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        uint32_ref (Uint32Refpts): uint32 numpy array used for indexing
          :py:class:`.Element`\ s in a lattice.

    Examples:

        >>> get_uint32_index(ring, at.Sextupole)
        array([ 21,  27,  35,  89,  97, 103], dtype=uint32)

        numpy array of indices of all :py:class:`.Sextupole`\ s

        >>> get_uint32_index(ring, at.End)
        array([121], dtype=uint32)

        numpy array([:pycode:`len(ring)+1`])

        >>> get_uint32_index(ring, at.checkattr('Frequency'))
        array([0], dtype=uint32)

        numpy array of indices of all elements having a 'Frequency'
        attribute
    """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        return uint32_refpts(refpts, len(ring), endpoint=endpoint, types=_typ2)

    return numpy.fromiter((i for i, el in enumerate(ring) if checkfun(el)),
                          dtype=numpy.uint32)


def bool_refpts(refpts: RefIndex, n_elements: int,
                endpoint: bool = True,
                types: str = _typ1) -> BoolRefpts:
    r"""Returns a :py:class:`bool` array of element indices, selecting ring
    elements.

    Parameters:
        refpts:     Element selector. *refpts* may be:

          #. an integer or a sequence of integers
             (0 indicating the first element),
          #. a sequence of booleans marking the selected elements,
          #. :py:obj:`None`, meaning empty selection,
          #. :py:obj:`.All`, meaning "all possible reference points",
          #. :py:obj:`.End`, selecting the end of the last element.
        n_elements: Length of the lattice
        endpoint:   if :py:obj:`True`, allow *n_elements* as a
          special index, referring to the end of the last element.
        types:      Allowed types

    Returns:
        bool_refs (BoolRefpts):  A bool numpy array used for indexing
          :py:class:`.Element`\ s in a lattice.
    """
    refs = numpy.ravel(refpts)
    stop = n_elements+1 if endpoint else n_elements
    if refpts is RefptsCode.All:
        return numpy.ones(stop, dtype=bool)
    elif refpts is RefptsCode.End:
        if not endpoint:
            raise IndexError('"End" index out of range')
        brefpts = numpy.zeros(stop, dtype=bool)
        brefpts[n_elements] = True
        return brefpts
    elif (refpts is None) or (refs.size == 0):
        return numpy.zeros(stop, dtype=bool)
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        diff = stop - refs.size
        if diff <= 0:
            return refs[:stop]
        else:
            return numpy.append(refs, numpy.zeros(diff, dtype=bool))
    elif numpy.issubdtype(refs.dtype, numpy.integer):
        brefpts = numpy.zeros(stop, dtype=bool)
        brefpts[refs] = True
        return brefpts
    else:
        raise _type_error(refpts, types)


# noinspection PyIncorrectDocstring
def get_bool_index(ring: Sequence[Element], refpts: Refpts,
                   endpoint: bool = True, regex: bool = False) -> BoolRefpts:
    # noinspection PyUnresolvedReferences, PyShadowingNames
    r"""Returns a bool array of element indices, selecting ring elements.

    Parameters:
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        endpoint:   if :py:obj:`True`, allow *len(ring)* as a
          special index, referring to the end of the last element.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        bool_refs (BoolRefpts):  A bool numpy array used for indexing
          :py:class:`.Element`\ s in a lattice.

    Examples:

        >>> refpts = get_bool_index(ring, at.Quadrupole)

        Returns a numpy array of booleans where all :py:class:`.Quadrupole`
        are :py:obj:`True`

        >>> refpts = get_bool_index(ring, "Q[FD]*")

        Returns a numpy array of booleans where all elements whose *FamName*
        matches "Q[FD]*" are :py:obj:`True`

        >>> refpts = get_bool_index(ring, at.checkattr('K', 0.0))

        Returns a numpy array of booleans where all elements whose *K*
        attribute is 0.0 are :py:obj:`True`

        >>> refpts = get_bool_index(ring, None)

        Returns a numpy array of *len(ring)+1* :py:obj:`False` values
    """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        return bool_refpts(refpts, len(ring), endpoint=endpoint, types=_typ2)

    boolrefs = numpy.fromiter(map(checkfun, ring), dtype=bool, count=len(ring))
    if endpoint:
        boolrefs = numpy.append(boolrefs, False)
    return boolrefs


def checkattr(attrname: str, attrvalue: Optional = None) \
        -> ElementFilter:
    # noinspection PyUnresolvedReferences
    r"""Checks the presence or the value of an attribute

    Returns a function to be used as an :py:class:`.Element` filter, which
    checks the presence or the value of an attribute of the
    provided :py:class:`.Element`.
    This function can be used to extract from a ring all elements
    having a given attribute.

    Parameters:
        attrname: Attribute name
        attrvalue: Attribute value. If absent, the returned function checks
          the presence of an *attrname* attribute. If present, the
          returned function checks if :pycode:`attrname == attrvalue`.

    Returns:
        checkfun (ElementFilter):   Element filter function

    Examples:

        >>> cavs = filter(checkattr('Frequency'), ring)

        Returns an iterator over all elements in *ring* that have a
        :pycode:`Frequency` attribute

        >>> elts = filter(checkattr('K', 0.0), ring)

        Returns an iterator over all elements in ring that have a
        :pycode:`K` attribute equal to 0.0
    """
    def testf(el):
        try:
            v = getattr(el, attrname)
            return (attrvalue is None) or (v == attrvalue)
        except AttributeError:
            return False

    return testf


def checktype(eltype: Union[type, Tuple[type, ...]]) -> ElementFilter:
    # noinspection PyUnresolvedReferences
    r"""Checks the type of an element

    Returns a function to be used as an :py:class:`.Element` filter, which
    checks the type of the provided :py:class:`.Element`.
    This function can be used to extract from a ring all elements
    having a given type.

    Parameters:
        eltype: Desired :py:class:`.Element` type

    Returns:
        checkfun (ElementFilter):   Element filter function

    Examples:

        >>> qps = filter(checktype(at.Quadrupole), ring)

        Returns an iterator over all quadrupoles in ring
    """
    return lambda el: isinstance(el, eltype)


def checkname(pattern: str, regex: bool = False) -> ElementFilter:
    # noinspection PyUnresolvedReferences
    r"""Checks the name of an element

    Returns a function to be used as an :py:class:`.Element` filter,
    which checks the name of the provided :py:class:`.Element`.
    This function can be used to extract from a ring all elements
    having a given name.

    Parameters:
        pattern: Desired :py:class:`.Element` name. Unix shell-style
          wildcards are supported (see :py:func:`fnmatch.fnmatch`)
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        checkfun (ElementFilter):   Element filter function

    Examples:

        >>> qps = filter(checkname('QF*'), ring)

        Returns an iterator over all with name starting with ``QF``.
    """
    if regex:
        rgx = re.compile(pattern)
        return lambda el: rgx.fullmatch(el.FamName)
    else:
        return lambda el: fnmatch(el.FamName, pattern)


def refpts_iterator(ring: Sequence[Element], refpts: Refpts,
                    regex: bool = False) -> Iterator[Element]:
    r"""Return an iterator over selected elements in a lattice

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        elem_iter (Iterator[Element]):  Iterator over the elements in *ring*
          selected by *refpts*.
    """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        refs = numpy.ravel(refpts)
        if refpts is RefptsCode.All:
            return (el for el in ring)
        elif refpts is RefptsCode.End:
            raise IndexError('"End" is not allowed for endpoint=False')
        elif (refpts is None) or (refs.size == 0):
            return iter(())
        elif numpy.issubdtype(refs.dtype, numpy.bool_):
            return compress(ring, refs)
        elif numpy.issubdtype(refs.dtype, numpy.integer):
            return (ring[i] for i in refs)
        else:
            raise _type_error(refpts, _typ2)

    return filter(checkfun, ring)


# noinspection PyUnusedLocal,PyIncorrectDocstring
def refpts_count(refpts: RefIndex, n_elements: int,
                 endpoint: bool = True,
                 types: str = _typ1) -> int:
    r"""Returns the number of reference points

    Parameters:
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element),
          #. a sequence of booleans marking the selected elements,
          #. :py:obj:`None`, meaning empty selection,
          #. :py:obj:`.All`, meaning "all possible reference points",
          #. :py:obj:`.End`, selecting the end of the last element.
        n_elements: Lattice length
        endpoint:   if :py:obj:`True`, allow *n_elements* as a
          special index, referring to the end of the last element.

    Returns:
        nrefs (int):  The number of reference points
    """
    refs = numpy.ravel(refpts)
    if refpts is RefptsCode.All:
        return n_elements+1 if endpoint else n_elements
    elif refpts is RefptsCode.End:
        if not endpoint:
            raise IndexError('"End" index out of range')
        return 1
    elif (refpts is None) or (refs.size == 0):
        return 0
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        return numpy.count_nonzero(refs)
    elif numpy.issubdtype(refs.dtype, numpy.integer):
        return len(refs)
    else:
        raise _type_error(refpts, types)


def _refcount(ring: Sequence[Element], refpts: Refpts,
              endpoint: bool = True, regex: bool = False) -> int:
    # noinspection PyUnresolvedReferences, PyShadowingNames
    r"""Returns the number of reference points

    Parameters:
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        endpoint:   if :py:obj:`True`, allow *len(ring)* as a
          special index, referring to the end of the last element.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        nrefs (int):  The number of reference points

    Examples:

        >>> refpts = ring.refcount(at.Sextupole)
        6

        Returns the number of :py:class:`.Sextupole`\ s in the lattice

        >>> refpts = ring.refcount(at.All)
        122

        Returns *len(ring)+1*

        >>> refpts = ring.refcount(at.All, endpoint=False)
        121

        Returns *len(ring)*
        """
    if isinstance(refpts, type):
        checkfun = checktype(refpts)
    elif callable(refpts):
        checkfun = refpts
    elif isinstance(refpts, Element):
        checkfun = checktype(type(refpts))
    elif isinstance(refpts, str):
        checkfun = checkname(refpts, regex=regex)
    else:
        return refpts_count(refpts, len(ring), endpoint=endpoint, types=_typ2)

    return len(list(filter(checkfun, ring)))


# noinspection PyUnusedLocal,PyIncorrectDocstring
def get_elements(ring: Sequence[Element], refpts: Refpts,
                 regex: bool = False) -> list:
    r"""Returns a list of elements selected by *key*.

    Deprecated: :pycode:`get_elements(ring, refpts)` is :pycode:`ring[refpts]`

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        elem_list (list):  list of :py:class:`.Element`\ s matching key
    """
    return list(refpts_iterator(ring, refpts, regex=regex))


def get_value_refpts(ring: Sequence[Element], refpts: Refpts,
                     attrname: str, index: Optional[int] = None,
                     regex: bool = False):
    r"""Extracts attribute values from selected
        lattice :py:class:`.Element`\ s.

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        attrname:   Attribute name
        index:      index of the value to retrieve if *attrname* is
          an array. If :py:obj:`None` the full array is retrieved
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        attrvalues: numpy Array of attribute values.
    """
    getf = getval(attrname, index=index)
    return numpy.array([getf(elem) for elem in refpts_iterator(ring, refpts,
                                                               regex=regex)])


def set_value_refpts(ring: Sequence[Element], refpts: Refpts,
                     attrname: str, attrvalues, index: Optional[int] = None,
                     increment: bool = False,
                     copy: bool = False, regex: bool = False):
    r"""Set the values of an attribute of an array of elements based on
    their refpts


    Parameters:
        ring:       Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        attrname:   Attribute name
        attrvalues: Attribute values
        index:      index of the value to set if *attrname* is
          an array. if :py:obj:`None`, the full array is replaced by
          *attrvalue*
        increment:  If :py:obj:`True`, *attrvalues* are added to the
          initial values.
        regex: Use regular expression for *refpts* string matching
          instead of Unix shell-style wildcards.
        copy:       If :py:obj:`False`, the modification is done in-place,

          If :py:obj:`True`, return a shallow copy of the lattice. Only the
          modified elements are copied.

          .. Caution::

             a shallow copy means that all non-modified
             elements are shared with the original lattice.
             Any further modification will affect both lattices.
    """
    setf = setval(attrname, index=index)
    if increment:
        attrvalues += get_value_refpts(ring, refpts,
                                       attrname, index=index,
                                       regex=regex)
    else:
        attrvalues = numpy.broadcast_to(attrvalues,
                                        (_refcount(ring, refpts,
                                                   regex=regex),))

    # noinspection PyShadowingNames
    @make_copy(copy)
    def apply(ring, refpts, values, regex):
        for elm, val in zip(refpts_iterator(ring, refpts, regex=regex), values):
            setf(elm, val)

    return apply(ring, refpts, attrvalues, regex)


def get_s_pos(ring: Sequence[Element], refpts: Refpts = All,
              regex: bool = False) -> Sequence[float]:
    # noinspection PyUnresolvedReferences
    r"""Returns the locations of selected elements

    Parameters:
        ring:       Lattice description
        refpts:     Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        s_pos:  Array of locations of the elements selected by *refpts*

    Example:

        >>> get_s_pos(ring, at.End)
        array([26.37428795])

        Position at the end of the last element: length of the lattice
        """
    # Positions at the end of each element.
    s_pos = numpy.cumsum([getattr(el, 'Length', 0.0) for el in ring])
    # Prepend position at the start of the first element.
    s_pos = numpy.concatenate(([0.0], s_pos))
    return s_pos[get_bool_index(ring, refpts, regex=regex)]


def rotate_elem(elem: Element, tilt: float = 0.0, pitch: float = 0.0,
                yaw: float = 0.0, relative: bool = False) -> None:
    r"""Set the tilt, pitch and yaw angle of an :py:class:`.Element`.
    The tilt is a rotation around the *s*-axis, the pitch is a
    rotation around the *x*-axis and the yaw is a rotation around
    the *y*-axis.

    A positive angle represents a clockwise rotation when
    looking in the direction of the rotation axis.

    The transformations are not all commmutative, the pitch and yaw
    are applied first and the tilt is always the last transformation
    applied. The element is rotated around its mid-point.

    If *relative* is :py:obj:`True`, the previous angle and shifts
    are rebuilt form the *R* and *T* matrix and incremented by the
    input arguments.

    The shift is always conserved regardless of the value of *relative*.

    The transformations are applied by changing the particle coordinates
    at the entrance of the element and restoring them at the end. Following
    the small angles approximation the longitudinal shift of the particle
    coordinates is neglected and the element length is unchanged.

    Parameters:
        elem:           Element to be tilted
        tilt:           Tilt angle [rad]
        pitch:          Pitch angle [rad]
        yaw:            Yaw angle [rad]
        relative:       If :py:obj:`True`, the rotation is added to the
          previous one
    """
    # noinspection PyShadowingNames
    def _get_rm_tv(le, tilt, pitch, yaw):
        tilt = numpy.around(tilt, decimals=15)
        pitch = numpy.around(pitch, decimals=15)
        yaw = numpy.around(yaw, decimals=15)
        ct, st = numpy.cos(tilt), numpy.sin(tilt)
        ap, ay = 0.5*le*numpy.tan(pitch), 0.5*le*numpy.tan(yaw)
        rr1 = numpy.asfortranarray(numpy.diag([ct, ct, ct, ct, 1.0, 1.0]))
        rr1[0, 2] = st
        rr1[1, 3] = st
        rr1[2, 0] = -st
        rr1[3, 1] = -st
        rr2 = rr1.T
        t1 = numpy.array([ay, numpy.sin(-yaw), -ap, numpy.sin(pitch), 0, 0])
        t2 = numpy.array([ay, numpy.sin(yaw), -ap, numpy.sin(-pitch), 0, 0])
        rt1 = numpy.eye(6, order='F')
        rt1[1, 4] = t1[1]
        rt1[3, 4] = t1[3]
        rt2 = numpy.eye(6, order='F')
        rt2[1, 4] = t2[1]
        rt2[3, 4] = t2[3]
        return rr1 @ rt1, rt2 @ rr2, t1, t2

    tilt0 = 0.0
    pitch0 = 0.0
    yaw0 = 0.0
    t10 = numpy.zeros(6)
    t20 = numpy.zeros(6)
    if hasattr(elem, 'R1') and hasattr(elem, 'R2'):
        rr10 = numpy.eye(6, order='F')
        rr10[:4, :4] = elem.R1[:4, :4]
        rt10 = rr10.T @ elem.R1
        tilt0 = numpy.arctan2(rr10[0, 2], rr10[0, 0])
        yaw0 = numpy.arcsin(-rt10[1, 4])
        pitch0 = numpy.arcsin(rt10[3, 4])
        _, _, t10, t20 = _get_rm_tv(elem.Length, tilt0, pitch0, yaw0)
    if hasattr(elem, 'T1') and hasattr(elem, 'T2'):
        t10 = elem.T1-t10
        t20 = elem.T2-t20
    if relative:
        tilt += tilt0
        pitch += pitch0
        yaw += yaw0

    r1, r2, t1, t2 = _get_rm_tv(elem.Length, tilt, pitch, yaw)
    elem.R1 = r1
    elem.R2 = r2
    elem.T1 = t1+t10
    elem.T2 = t2+t20


def tilt_elem(elem: Element, rots: float, relative: bool = False) -> None:
    r"""Set the tilt angle :math:`\theta` of an :py:class:`.Element`

    The rotation matrices are stored in the :pycode:`R1` and :pycode:`R2`
    attributes.

    :math:`R_1=\begin{pmatrix} cos\theta & sin\theta \\
    -sin\theta & cos\theta \end{pmatrix}`,
    :math:`R_2=\begin{pmatrix} cos\theta & -sin\theta \\
    sin\theta & cos\theta \end{pmatrix}`

    Parameters:
        elem:           Element to be tilted
        rots:           Tilt angle :math:`\theta` [rd]. *rots* > 0 corresponds
          to a corkscrew rotation of the element looking in the direction of
          the beam
        relative:       If :py:obj:`True`, the rotation is added to the
          existing one
    """
    rotate_elem(elem, tilt=rots, relative=relative)


def shift_elem(elem: Element, deltax: float = 0.0, deltaz: float = 0.0,
               relative: bool = False) -> None:
    r"""Sets the transverse displacement of an :py:class:`.Element`

    The translation vectors are stored in the :pycode:`T1` and :pycode:`T2`
    attributes.

    Parameters:
        elem:           Element to be shifted
        deltax:         Horizontal displacement [m]
        deltaz:         Vertical displacement [m]
        relative:       If :py:obj:`True`, the translation is added to the
          existing one
    """
    tr = numpy.array([deltax, 0.0, deltaz, 0.0, 0.0, 0.0])
    if relative and hasattr(elem, 'T1') and hasattr(elem, 'T2'):
        elem.T1 -= tr
        elem.T2 += tr
    else:
        elem.T1 = -tr
        elem.T2 = tr


def set_rotation(ring: Sequence[Element], tilts=0.0,
                 pitches=0.0, yaws=0.0, relative=False) -> None:
    r"""Sets the tilts of a list of elements.

    Parameters:
        ring:           Lattice description
        tilts:          Sequence of tilt values as long as ring or
          scalar tilt value applied to all elements, default=0
        pitches:          Sequence of pitch values as long as ring or
          scalar tilt value applied to all elements, default=0
        yaws:          Sequence of yaw values as long as ring or
          scalar tilt value applied to all elements, default=0
        relative:       If :py:obj:`True`, the rotations are added to the
          existing ones
    """
    tilts = numpy.broadcast_to(tilts, (len(ring),))
    pitches = numpy.broadcast_to(pitches, (len(ring),))
    yaws = numpy.broadcast_to(yaws, (len(ring),))
    for el, tilt, pitch, yaw in zip(ring, tilts, pitches, yaws):
        rotate_elem(el, tilt=tilt, pitch=pitch, yaw=yaw, relative=relative)


def set_tilt(ring: Sequence[Element], tilts, relative=False) -> None:
    r"""Sets the tilts of a list of elements.

    Parameters:
        ring:           Lattice description
        tilts:          Sequence of tilt values as long as ring or
          scalar tilt value applied to all elements
        relative:       If :py:obj:`True`, the rotation is added to the
          existing one
    """
    tilts = numpy.broadcast_to(tilts, (len(ring),))
    for el, tilt in zip(ring, tilts):
        tilt_elem(el, tilt, relative=relative)


def set_shift(ring: Sequence[Element], dxs, dzs, relative=False) -> None:
    r"""Sets the translations of a list of elements.

    Parameters:
        ring:           Lattice description
        dxs:            Sequence of horizontal displacements values as long as
          ring or scalar value applied to all elements [m]
        dzs:            Sequence of vertical displacements values as long as
          ring or scalar value applied to all elements [m]
        relative:       If :py:obj:`True`, the displacement is added to the
          existing one
    """
    dxs = numpy.broadcast_to(dxs, (len(ring),))
    dzs = numpy.broadcast_to(dzs, (len(ring),))
    for el, dx, dy in zip(ring, dxs, dzs):
        shift_elem(el, dx, dy, relative=relative)


def get_geometry(ring: List[Element],
                 refpts: Refpts = All,
                 start_coordinates: Tuple[float, float, float] = (0, 0, 0),
                 centered: bool = False,
                 regex: bool = False
                 ):
    # noinspection PyShadowingNames
    r"""Compute the 2D ring geometry in cartesian coordinates

    Parameters:
        ring:               Lattice description.
        refpts:     Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        start_coordinates:  *x*, *y*, *angle* at starting point. *x*
          and *y* are ignored if *centered* is :py:obj:`True`.
        centered:           if :py:obj:`True` the coordinates origin is the
          center of the ring.
        regex: Use regular expression for *refpts* string matching instead of
          Unix shell-style wildcards.

    Returns:
        geomdata:           recarray containing, x, y, angle.
        radius:             machine radius at the beginning of the lattice.

            .. attention::
               This radius is different from the radius usually defined as
               :math:`C/2\pi`

    Example:

       >>> geomdata, radius = get_geometry(ring)
    """

    geom_dtype = [("x", numpy.float64),
                  ("y", numpy.float64),
                  ("angle", numpy.float64)]
    boolrefs = get_bool_index(ring, refpts, endpoint=True, regex=regex)
    nrefs = refpts_count(boolrefs, len(ring))
    geomdata = numpy.recarray((nrefs, ), dtype=geom_dtype)
    xx = numpy.zeros(len(ring)+1)
    yy = numpy.zeros(len(ring)+1)
    angle = numpy.zeros(len(ring)+1)
    x0, y0, t0 = start_coordinates
    x, y = 0.0, 0.0
    t = t0

    xx[0] = x
    yy[0] = y
    angle[0] = t
    for ind, el in enumerate(ring):
        ll = el.Length
        if isinstance(el, Dipole) and el.BendingAngle != 0:
            ang = 0.5 * el.BendingAngle
            ll *= numpy.sin(ang)/ang
        else:
            ang = 0.0
        t -= ang
        x += ll * numpy.cos(t)
        y += ll * numpy.sin(t)
        t -= ang
        xx[ind+1] = x
        yy[ind+1] = y
        angle[ind+1] = t

    dff = (t + _GEOMETRY_EPSIL) % (2.0 * numpy.pi) - _GEOMETRY_EPSIL
    if abs(dff) < _GEOMETRY_EPSIL:
        xcenter = numpy.mean(xx)
        ycenter = numpy.mean(yy)
    elif abs(dff-numpy.pi) < _GEOMETRY_EPSIL:
        xcenter = 0.5*x
        ycenter = 0.5*y
    else:
        num = numpy.cos(t)*x + numpy.sin(t)*y
        den = numpy.sin(t-t0)
        xcenter = -num*numpy.sin(t0)/den
        ycenter = num*numpy.cos(t0)/den
    radius = numpy.sqrt(xcenter*xcenter + ycenter*ycenter)
    if centered:
        xx -= xcenter
        yy -= ycenter
    else:
        xx += x0
        yy += y0
    geomdata["x"] = xx[boolrefs]
    geomdata["y"] = yy[boolrefs]
    geomdata["angle"] = angle[boolrefs]
    return geomdata, radius



import numpy
import matplotlib.pyplot as plt
import matplotlib.cm as cmap
import at
from at.collective import BeamLoadingElement, add_beamloading, BLMode
from mpi4py import MPI
from at.tracking.utils import get_bunches_std_mean
from at.constants import qe
from at.physics import harmonic_analysis


def analytical_qs(ring, I):

    E0 = ring.energy * qe
    alpha_c = at.get_mcf(ring.radiation_off(copy=True))
    omega_rf = 2 * numpy.pi * ring.get_rf_frequency()
    Vc = ring.get_rf_voltage()

    envel = at.envelope_parameters(ring.radiation_on(copy=True))
    U0 = envel.U0 * qe

    Q0 = bl_elem.Qfactor * (1 + beta)
    Rsh = bl_elem.Rshunt * (1 + beta)  # Difference in definition

    phi_s = numpy.pi - numpy.arcsin(U0 / (qe * Vc))  # sync. phase in radian
    psi = numpy.arctan(-2 * Rsh * I * numpy.cos(numpy.pi - phi_s) /
                         (Vc * (1 + beta)))  # tuning angle
                         
    omega_res = (omega_rf /
                 (1 - 2 * Rsh * I * (numpy.cos(numpy.pi - phi_s)) /
                  (2 * Q0 * Vc)))  # resonant freq.

    h = ring.harmonic_number  # harmonic number
    omega0 = omega_rf / h  # rev. freq.
    T0 = 2 * numpy.pi / omega0  # rev. time

    omega_s0 = numpy.sqrt(qe * Vc * omega_rf * alpha_c * numpy.cos(numpy.pi - phi_s) /
                          (E0 * T0))  # Synch. freq. for a single particle

    K = alpha_c * qe * I / (E0 * T0) * Rsh / (1 + beta)

    # Synch. freq. with beam loading
    tilde_omega_s = numpy.sqrt(omega_s0**2 +
                               K * omega_rf * numpy.sin(2 * psi) + 1j * 0)
    return tilde_omega_s / omega0, omega_s0/omega0


comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

print(size, rank)

ring = at.load_lattice('../../../machine_data/esrf.m')
ring.enable_6d(cavity_pass='RFCavityPass')
ring.set_rf_frequency()


# In fact, the following resonator parameters are for the EBS machine
# Not the old ESRF SR, however for benchmarking this doesn't matter
beta = 2.8
q0 = 37500  # unloaded
qfactor = q0 / (1 + beta)  # loaded
rshunt = 145 * qfactor * 13  # loaded
o6, _ = ring.find_orbit6()

ring.set_rf_voltage(8.e6)

Nbunches = 16
ring.set_fillpattern(Nbunches)

_, fring = at.fast_ring(ring)
fring.pop(-1)  # drop diffusion element

# Here we specify whether we want to use PHASOR or WAKE
# beam loading models.
mode = 'WAKE'
if mode == 'WAKE':
    blm = BLMode.WAKE
else:
    blm = BLMode.PHASOR

# Npart must be at least Nbunches per core
Npart = Nbunches

# Now we give the fring and convert it
# into a beam loaded cavity.
add_beamloading(fring, qfactor, rshunt, Nslice=1,
                Nturns=50, blmode=blm,
                VoltGain=0.1, PhaseGain=0.1)

bl_elem = fring[at.get_refpts(fring, BeamLoadingElement)[0]]

# Specify some simulation parameters
kickTurn = 500
Nturns = 2**14 + kickTurn
current = 150e-3
fring.beam_current = current

part = numpy.zeros((6, Npart))
part = (part.T + o6).T

if rank == 0:

    print('\n')
    print('Current:', current)

    z_all = numpy.zeros((Nturns, Nbunches))
    dp_all = numpy.zeros((Nturns, Nbunches))


# Here it should be considered that there are 2 ways to run
# simulations in pyat. Either with ring.track(part, nturns=Nturns)
# or with for i in np.arange(Nturns): ring.track(part, nturns=1).
# With the former, you should use the BeamMoments element to acquire
# the means and stds of each bunch turn by turn when using MPI.
# However, when you want to kick after a certain turn, you have 1
# of 2 choices. Either you use a BeamMoments element, and split the 
# tracking into 2 (before and after). You can then concatenate the
# the results. Or you can use the code below to gather all particles
# yourself with the MPICOMM after each turn.

for i in numpy.arange(Nturns):
    # Apply a kick to ensure the coherent tune is large
    if i == kickTurn:
        part[4, :] += 3e-3

    fring.track(part, nturns=1, refpts=None, in_place=True)

    # Gather particles over all cores (compatible with MPI on or off)
    allPartsg = comm.gather(part)
    if rank == 0:
        allPartsg = numpy.concatenate(allPartsg, axis=1)
        stds, means = get_bunches_std_mean(allPartsg, Nbunches)

        dp_all[i] = numpy.array([x[4] for x in means])
        z_all[i] = numpy.array([x[5] for x in means])
        if i % 1000 == 0:
            # Print the turn number and number of lost particles
            print(i, numpy.sum(numpy.isnan(allPartsg)))


# Now we compute the tune of each bunch, and compare with the analytical
if rank == 0:
    qscoh = numpy.zeros(Nbunches)
    for ib in numpy.arange(Nbunches):
        dp_dat = dp_all[kickTurn:, ib] - numpy.mean(dp_all[kickTurn:, ib])
        qs = harmonic_analysis.get_tunes_harmonic(dp_dat,
                                                  num_harmonics=20,
                                                  fmin=1e-5, fmax=0.1)
        qscoh[ib] = qs

    qs_mn, qs_std = numpy.array([numpy.mean(qscoh), numpy.std(qscoh)])

    qs_theory, qs_zerocurrent = analytical_qs(ring, current)

    print('Analytical:', numpy.real(qs_theory))
    print('Simulated:', qs_mn, 'pm', qs_std)

    freq = numpy.fft.rfftfreq(Nturns - kickTurn)
    for i in numpy.arange(Nbunches):
        fftdat = numpy.abs(numpy.fft.rfft(dp_all[kickTurn:, i]))
        if i == 0:
            plt.semilogy(freq, fftdat,
                         color=cmap.jet(float(i + 1) / Nbunches),
                         label='Bunch by Bunch FFT')
        else:
            plt.semilogy(freq, fftdat,
                         color=cmap.jet(float(i + 1) / Nbunches))

    plt.xlim([0, 8e-3])
    plt.xlabel('Tune')
    plt.ylabel('FFT Amp [A.U.]')
    plt.axvline(numpy.mean(qscoh), label='Coherent Tune',
                linestyle='dashed', color='k')
    plt.axvline(numpy.real(qs_theory), label='Analytical Beam Loaded Tune',
                linestyle='dashed', color='r')
    plt.axvline(qs_zerocurrent, label='Zero Current Tune',
                linestyle='dashed', color='b')
    plt.legend()
    plt.show()



import numpy
'''
This is an example of a python passmethod, here the Drift
passmethod. This is a simplified version that does not check
aperture.

The python passmethods has to be in the python path or in 
at.integratorsand contain a function trackFunction()
'''


def trackFunction(rin, elem=None):
    scaling = elem.FieldScaling
    for i in range(0, len(rin), 6):
        rin[i+1] /= scaling
        rin[i+3] /= scaling
        rin[i+4] = (rin[i+4] + 1.0 - scaling) / scaling
    return rin



import numpy
import pytest

from at import elements
from at import lattice_pass, internal_lpass
from at import lattice_track


@pytest.mark.parametrize("func", (lattice_track, lattice_pass, internal_lpass))
@pytest.mark.parametrize("input_dim", [(0,), (5,), (7,), (1, 1), (6, 1, 1)])
def test_lattice_track_raises_AssertionError_if_rin_incorrect_shape(input_dim, func):
    rin = numpy.zeros(input_dim)
    lattice = []
    with pytest.raises(AssertionError):
        func(lattice, rin)


@pytest.mark.parametrize("func", (lattice_track, lattice_pass, internal_lpass))
def test_multiple_particles_lattice_track(func):
    lattice = [elements.Drift("Drift", 1.0)]
    rin = numpy.zeros((6, 2))
    rin[0, 0] = 1e-6  # particle one offset in x
    rin[2, 1] = 1e-6  # particle two offset in y
    r_original = numpy.copy(rin)
    r_out = func(lattice, rin, nturns=2)
    if isinstance(r_out, tuple):
        r_out, *_ = r_out
    # particle position is not changed passing through the drift
    numpy.testing.assert_equal(r_original[:, 0], r_out[:, 0, 0, 0])
    numpy.testing.assert_equal(r_original[:, 0], r_out[:, 0, 0, 1])
    numpy.testing.assert_equal(r_original[:, 1], r_out[:, 1, 0, 0])
    numpy.testing.assert_equal(r_original[:, 1], r_out[:, 1, 0, 1])


@pytest.mark.parametrize("func", (lattice_track, lattice_pass))
def test_lattice_convert_to_list_if_incorrect_type(func):
    lattice = numpy.array([elements.Drift("Drift", 1.0)])
    rin = numpy.zeros((6, 2))
    rin[0, 0] = 1e-6
    r_original = numpy.copy(rin)
    r_out = func(lattice, rin, 1)
    if isinstance(r_out, tuple):
        r_out, *_ = r_out
    numpy.testing.assert_equal(r_original, r_out.reshape(6, 2))


def test_in_place_flag_tracking():
    lattice = numpy.array([elements.Drift("Drift", 1.0)])
    rin = numpy.zeros((6, 1))
    rin[1] = 1e-6
    r_original = numpy.copy(rin)
    lattice_track(lattice, rin)
    numpy.testing.assert_equal(r_original, rin.reshape(6, 1))
    rout, *_ = lattice_track(lattice, rin, in_place=True)
    numpy.testing.assert_equal(rin, rout.reshape(6, 1))



"""AT plotting functions"""
from __future__ import annotations
from at.lattice import Lattice, axis_
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
import numpy
from numpy import ndarray
from math import sqrt


# Function to compute and plot acceptance
def plot_acceptance(ring: Lattice, planes, *args, **kwargs):
    # noinspection PyUnresolvedReferences
    r"""Plots the acceptance

    Computes the acceptance at repfts observation points using
    :py:func:`.get_acceptance` and plots the tracked
    and survived particles, and the acceptance boundary.

    Parameters:
        ring:           Lattice definition
        planes:         max. dimension 2, Plane(s) to scan for the acceptance.
          Allowed values are: *'x'*, *'xp'*, *'y'*,
          *'yp'*, *'dp'*, *'ct'*

    Keyword Args:
        acceptance (tuple[ndarray, ndarray, ndarray]): tuple containing
          pre-computed acceptance *(boundary, survived, grid)*
        npoints:        (len(planes),) array: number of points in each
          dimension
        amplitudes:     (len(planes),) array: set the search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step
        nturns (int):       Number of turns for the tracking
        obspt (Refpts):    Observation points. Default: start of the machine
        dp (float):         Static momentum offset
        offset:             Initial orbit. Default: closed orbit
        bounds:             Defines the tracked range: range=bounds*amplitude.
          It can be used to select quadrants. For example, default values are:

          * :py:attr:`.GridMode.CARTESIAN`: ((-1, 1), (0, 1))
          * :py:attr:`GridMode.RADIAL/RECURSIVE <.GridMode.RADIAL>`: ((0, 1),
            (:math:`\pi`, 0))
        grid_mode (GridMode):   Defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp (bool):      Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multiprocessing is not
          enabled, *grid_mode* is forced to :py:attr:`.GridMode.RECURSIVE`
          (most efficient in single core)
        verbose (bool):     Print out some information
        divider (int):      Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero:
        start_method (str): Python multiprocessing start method. The default
          :py:obj:`None` uses the python default that is considered safe.
          Available parameters: *'fork'*, *'spawn'*, *'forkserver'*.
          The default for linux is *'fork'*, the default for macOS and
          Windows is *'spawn'*. *'fork'* may be used for macOS to speed up
          the calculation or to solve runtime errors, however  it is
          considered unsafe.

    Returns:
        boundary:   (2,n) array: 2D acceptance
        tracked:    (2,n) array: Coordinates of tracked particles
        survived:   (2,n) array: Coordinates of surviving particles

    Example:
        >>> ring.plot_acceptance(planes, npoints, amplitudes)
        >>> plt.show()
    """
    obspt = kwargs.pop('obspt', None)
    block = kwargs.pop('block', False)
    acceptance = kwargs.pop('acceptance', None)
    if obspt is not None:
        assert numpy.isscalar(obspt), 'Scalar value needed for obspt'
    kwargs['refpts'] = obspt
    if acceptance is None:
        boundary, survived, grid = ring.get_acceptance(planes, *args, **kwargs)
    else:
        boundary, survived, grid = acceptance
    plt.figure()
    plt.plot(*grid, '.', label='Tracked particles')
    plt.plot(*survived, '.', label='Survived particles')
    if len(planes) == 1:
        pl0 = axis_(planes[0])
        plt.plot(boundary, numpy.zeros(2), label='Acceptance')
        plt.title('1D {0} acceptance'.format(pl0['label']))
        plt.xlabel('{0}{1}'.format(pl0['label'], pl0['unit']))
    else:
        pl0, pl1 = axis_(planes)
        plt.plot(*boundary, label='Acceptance')
        plt.title('2D {0}-{1} acceptance'.format(pl0['label'], pl1['label']))
        plt.xlabel('{0}{1}'.format(pl0['label'], pl0['unit']))
        plt.xlabel('{0}{1}'.format(pl1['label'], pl1['unit']))
    plt.legend()
    plt.show(block=block)
    return boundary, survived, grid


def plot_geometry(ring: Lattice,
                  start_coordinates: tuple[float, float, float] = (0, 0, 0),
                  centered: bool = False, ax: Axes = None, **kwargs):
    """Compute and plot the 2D ring geometry in cartesian coordinates.

    Parameters:
        ring: Lattice description
        start_coordinates: x,y,angle at starting point
        centered: it True the coordinates origin is the center of the ring
        ax: axes for the plot. If not given, new axes are created

    Keyword arguments are forwarded to the underlying
    :py:func:`~matplotlib.pyplot.plot` function

    Returns:
        geomdata: recarray containing, x, y, angle
        radius: machine radius
        ax: plot axis

    Example:
        >>> ring.plot_geometry()
    """
    if not ax:
        fig, ax = plt.subplots()
    geom, radius = ring.get_geometry(start_coordinates=start_coordinates,
                                     centered=centered)
    ax.plot(geom['x'], geom['y'], 'o:',
            linewidth=kwargs.pop('linewidth', 0.5),
            markersize=kwargs.pop('markersize', 2),
            **kwargs)
    ax.set_xlabel('x [m]')
    ax.set_ylabel('y [m]')
    ax.set_aspect('equal', 'box')
    return geom, radius, ax


def plot_sigma(sigma, axis: tuple[str, str] = ('x', 'xp'), scale: float = 1.0,
               ax: Axes = None, **kwargs):
    r"""Plot the projection of the phase space defined by a
    :math:`\Sigma`-matrix on the selected plane.

    Arguments:
        sigma:  :math:`\Sigma`-matrix
        axis:   tuple if indices defining the plane of the :math:`\Sigma`
          projection. Allowed values are: *'x'*, *'xp'*, *'y'*,
          *'yp'*, *'dp'*, *'ct'*. Default: (*'x'*, *'xp'*)
        scale:  Scaling factor for the emittance
        ax: axes for the plot. If not given, new axes are created

    Keyword arguments are forwarded to the underlying
    :py:func:`~matplotlib.pyplot.plot` function
    """
    if not ax:
        fig, ax = plt.subplots()
    ax1, ax2 = axis_(axis)
    axid = axis_(axis, key='index')
    sig22 = sigma[numpy.ix_(axid, axid)]
    eps = sqrt(sig22[0, 0] * sig22[1, 1] - sig22[1, 0] * sig22[0, 1])
    sigx = sqrt(sig22[0, 0])
    tr = numpy.array([[sigx, 0.0],
                      [sig22[0, 1] / sigx, eps / sigx]])
    loop = 2.0 * numpy.pi * numpy.arange(0.0, 1.0, 0.001)
    normcoord = numpy.vstack((numpy.cos(loop), numpy.sin(loop)))
    coord = tr @ normcoord
    line = ax.plot(scale*coord[0, :], scale*coord[1, :], **kwargs)
    ax.set_title('{0}-{1} phase space'.format(ax1['label'], ax2['label']))
    ax.set_xlabel('{0}{1}'.format(ax1['label'], ax1['unit']))
    ax.set_ylabel('{0}{1}'.format(ax2['label'], ax2['unit']))
    return line


Lattice.plot_acceptance = plot_acceptance
Lattice.plot_geometry = plot_geometry



"""AT plotting functions"""
try:
    # noinspection PyPackageRequirements
    import matplotlib
    from .synopt import *
    from .generic import *
    from .specific import *
    from .standalone import *
except (ImportError, RuntimeError) as exc:
    print(exc)
    print('Plotting is disabled')



import pytest
from numpy.testing import assert_allclose as assert_close

from at import linopt2, linopt4, linopt6, get_optics


@pytest.mark.parametrize("lattice", ["dba_lattice", "hmba_lattice"])
def test_linopt6_norad(request, lattice):
    """Compare the results of linopt2 and linopt6 in 4d"""
    lattice = request.getfixturevalue(lattice)
    refpts = range(len(lattice) + 1)
    ld02, rd2, ld2 = linopt2(lattice, refpts, get_w=True)
    ld06, rd6, ld6 = linopt6(lattice, refpts, get_w=True)
    assert_close(rd2.tune, rd6.tune, atol=1e-10, rtol=0)
    assert_close(rd2.chromaticity, rd6.chromaticity, atol=1e-10, rtol=0)

    for field in ["s_pos", "closed_orbit", "dispersion", "alpha", "beta", "mu"]:
        assert_close(ld2[field], ld6[field], atol=1e-10, rtol=0, err_msg=field)
    assert_close(ld2.W, ld6.W, atol=1e-6, rtol=0)


@pytest.mark.parametrize("lattice", ["hmba_lattice"])
def test_linopt6_rad(request, lattice):
    """Compare the results with and without radiation"""
    lattice = request.getfixturevalue(lattice)
    refpts = range(len(lattice) + 1)
    # Turn cavity ON, without radiation
    radlattice = lattice.radiation_on(dipole_pass=None, copy=True)
    ld04, rd4, ld4 = linopt6(lattice, refpts, get_w=True)
    ld06, rd6, ld6 = linopt6(radlattice, refpts, get_w=True)

    assert_close(rd4.tune, rd6.tune[:2], atol=1e-10, rtol=0)
    assert_close(rd4.chromaticity, rd6.chromaticity[:2], atol=1e-8, rtol=0)

    for field in ["s_pos", "closed_orbit", "dispersion", "alpha", "beta"]:
        assert_close(ld4[field], ld6[field], atol=1.0e-8, rtol=0, err_msg=field)
    assert_close(ld4.mu, ld6.mu[:, :2], atol=1.0e-8, rtol=0)
    assert_close(ld4.W, ld6.W, atol=1e-6, rtol=2e-5)


@pytest.mark.parametrize("dp", (-0.005, 0.0, 0.005))
@pytest.mark.parametrize("lattice", ["dba_lattice", "hmba_lattice"])
@pytest.mark.parametrize("method", [linopt2, linopt4, linopt6])
def test_linopt6_line(request, lattice, dp, method):
    """Run a ring as a transfer line and compare with the ring results"""
    lattice = request.getfixturevalue(lattice)
    refpts = lattice.uint32_refpts(range(len(lattice) + 1))
    ld04, rd4, ld4 = get_optics(lattice, refpts, dp=dp, method=method)

    twin = dict(
        alpha=ld04.alpha,
        beta=ld04.beta,
        dispersion=ld04.dispersion,
        closed_orbit=ld04.closed_orbit,
    )
    # twin = ld04
    tr04, bd4, tr4 = get_optics(lattice, refpts, dp=dp, twiss_in=twin, method=method)
    for field in ["s_pos", "closed_orbit", "dispersion", "alpha", "beta", "mu"]:
        assert_close(ld04[field], tr04[field], atol=1.0e-10, rtol=0, err_msg=field)
        assert_close(ld4[field], tr4[field], atol=1.0e-8, rtol=0, err_msg=field)



"""Helper functions for axis and plane descriptions"""

from __future__ import annotations
from typing import Optional, Union

# For sys.version_info.minor < 9:
from typing import Tuple

AxisCode = Union[str, int, slice, None, type(Ellipsis)]
AxisDef = Union[AxisCode, Tuple[AxisCode, AxisCode]]

_axis_def = dict(
    x=dict(index=0, label="x", unit=" [m]"),
    px=dict(index=1, label=r"$p_x$", unit=" [rad]"),
    y=dict(index=2, label="y", unit=" [m]"),
    py=dict(index=3, label=r"$p_y$", unit=" [rad]"),
    dp=dict(index=4, label=r"$\delta$", unit=""),
    ct=dict(index=5, label=r"$\beta c \tau$", unit=" [m]"),
)
for xk, xv in [it for it in _axis_def.items()]:
    xv["code"] = xk
    _axis_def[xv["index"]] = xv
    _axis_def[xk.upper()] = xv
_axis_def["delta"] = _axis_def["dp"]
_axis_def["xp"] = _axis_def["px"]  # For backward compatibility
_axis_def["yp"] = _axis_def["py"]  # For backward compatibility
_axis_def["s"] = _axis_def["ct"]
_axis_def["S"] = _axis_def["ct"]
_axis_def[None] = dict(index=None, label="", unit="", code=":")
_axis_def[Ellipsis] = dict(index=Ellipsis, label="", unit="", code="...")

_plane_def = dict(
    x=dict(index=0, label="x", unit=" [m]"),
    y=dict(index=1, label="y", unit=" [m]"),
    z=dict(index=2, label="z", unit=""),
)
for xk, xv in [it for it in _plane_def.items()]:
    xv["code"] = xk
    _plane_def[xv["index"]] = xv
    _plane_def[xk.upper()] = xv
_plane_def["h"] = _plane_def["x"]
_plane_def["v"] = _plane_def["y"]
_plane_def["H"] = _plane_def["x"]
_plane_def["V"] = _plane_def["y"]
_plane_def[None] = dict(index=None, label="", unit="", code=":")
_plane_def[Ellipsis] = dict(index=Ellipsis, label="", unit="", code="...")


def _descr(dd: dict, arg: AxisDef, key: Optional[str] = None):
    if isinstance(arg, tuple):
        return tuple(_descr(dd, a, key=key) for a in arg)
    else:
        try:
            descr = dd[arg]
        except (TypeError, KeyError):
            descr = dict(index=arg, code=arg, label="", unit="")
        if key is None:
            return descr
        else:
            return descr[key]


def axis_(axis: AxisDef, key: Optional[str] = None):
    r"""Return axis descriptions

    Parameters:
        axis:           axis code or tuple of axis codes selecting axes in the
          standard AT coordinate system. codes for the 6 axes are:

          0, 'x', 'X'

          1, 'px', PX'

          2, 'y', 'Y'

          3, 'py', 'PY'

          4, 'dp', 'DP', 'delta'

          5, 'ct', 'CT', 's', 'S'

          :py:obj:`None` and :py:obj:`Ellipsis` select all axes

        key:            key in the coordinate descriptor dictionary,
          selecting the desired information. One of :

          'index'
            index in the standard AT coordinate vector
          'code'
            string representation
          'label'
            label for plot annotation
          'unit'
            coordinate unit
          :py:obj:`None`
            entire description dictionary

    Returns:
        descr : value or tuple[values]

    Examples:

        >>> axis_(('x','dp'), key='index')
        (0, 4)

        returns the indices in the standard coordinate vector

        >>> dplabel = axis_('dp', key='label')
        >>> print(dplabel)
        $\delta$

        returns the coordinate label for plot annotation

        >>> axis_((0,'dp'))
        ({'plane': 0, 'label': 'x', 'unit': ' [m]', 'code': 'x'},
         {'plane': 4, 'label': '$\\delta$', 'unit': '', 'code': 'dp'})

        returns the entire description directories

    """
    return _descr(_axis_def, axis, key=key)


def plane_(plane: AxisDef, key: Optional[str] = None):
    r"""Return plane descriptions

    Parameters:
        plane:          plane code or tuple of plane codes selecting planes.
          codes for the 3 planes are:

          0, 'x', 'X', 'h', 'H' for horizontal plane,

          1, 'y', 'Y', 'v', 'V' for vertical plane,

          2, 'z', 'Z'   for the longitudinal plane

          :py:obj:`None`, slice(None) and :py:obj:`Ellipsis` selects all planes
        key:            key in the plane descriptor dictionary,
          selecting the desired information. One of :

          'plane'
            plane in the optics data
          'code'
            string representation
          'label'
            label for plot annotation
          'unit'
            coordinate unit
          :py:obj:`None`
            entire description dictionary

    Returns:
        descr : value or tuple[values]

    Examples:

        >>> plane_('v', key='index')
        1

        returns the indices in the standard coordinate vector

        >>> plane_(('x','y'))
        ({'plane': 0, 'label': 'h', 'unit': ' [m]', 'code': 'h'},
         {'plane': 1, 'label': 'v', 'unit': ' [m]', 'code': 'v'})

        returns the entire description directories

    """
    return _descr(_plane_def, plane, key=key)



"""Lattice object

The methods implemented in this module are internal to the 'lattice' package.
This is necessary to ensure that the 'lattice' package is independent of other
AT packages.

Other Lattice methods are implemented in other AT packages and are available
as soon as the package is imported. The 'tracking' and 'physics' packages are
automatically imported.

As an example, see the at.physics.orbit module
"""
from __future__ import annotations

__all__ = ['Lattice', 'Filter', 'type_filter', 'params_filter', 'lattice_filter',
           'elem_generator', 'no_filter']

import sys
import copy
import math
from typing import Optional, Union
if sys.version_info.minor < 9:
    from typing import Callable, Iterable, Generator
    SupportsIndex = int
else:
    from typing import SupportsIndex
    from collections.abc import Callable, Iterable, Generator
from warnings import warn

import numpy

from . import elements as elt
from .elements import Element
from .particle_object import Particle
from .utils import AtError, AtWarning, Refpts
from .utils import get_s_pos, get_elements, get_value_refpts, set_value_refpts
# noinspection PyProtectedMember
from .utils import get_uint32_index, get_bool_index, _refcount, Uint32Refpts
from .utils import refpts_iterator, checktype, set_shift, set_tilt, get_geometry
from ..constants import clight, e_mass

_TWO_PI_ERROR = 1.E-4
Filter = Callable[..., Iterable[Element]]

_DEFAULT_PASS = {
    False: (
        ('cavity_pass', elt.RFCavity, 'auto'),
        ('dipole_pass', elt.Dipole, 'auto'),
        ('quadrupole_pass', elt.Quadrupole, 'auto'),
        ('wiggler_pass', elt.Wiggler, 'auto'),
        ('sextupole_pass', elt.Sextupole, 'auto'),
        ('octupole_pass', elt.Octupole, 'auto'),
        ('multipole_pass', elt.Multipole, 'auto'),
        ('collective_pass', elt.Collective, 'auto'),
        ('diffusion_pass', elt.QuantumDiffusion, 'auto'),
        ('energyloss_pass', elt.EnergyLoss, 'auto'),
        ('simplequantdiff_pass', elt.SimpleQuantDiff, 'auto'),
        ('simpleradiation_pass', elt.SimpleRadiation, 'auto')
    ),
    True: (
        ('cavity_pass', elt.RFCavity, 'auto'),
        ('dipole_pass', elt.Dipole, 'auto'),
        ('quadrupole_pass', elt.Quadrupole, 'auto'),
        ('wiggler_pass', elt.Wiggler, 'auto'),
        ('sextupole_pass', elt.Sextupole, None),
        ('octupole_pass', elt.Octupole, None),
        ('multipole_pass', elt.Multipole, None),
        ('collective_pass', elt.Collective, 'auto'),
        ('diffusion_pass', elt.QuantumDiffusion, 'auto'),
        ('energyloss_pass', elt.EnergyLoss, 'auto'),
        ('simplequantdiff_pass', elt.SimpleQuantDiff, 'auto'),
        ('simpleradiation_pass', elt.SimpleRadiation, 'auto')
    )
}

# Don't warn on floating-point errors
numpy.seterr(divide='ignore', invalid='ignore')


# noinspection PyAttributeOutsideInit
class Lattice(list):
    """Lattice object.

    An AT lattice is a sequence of AT elements. In addition to :py:class:`list`
    methods, :py:class:`Lattice` supports
    `extended indexing <https://numpy.org/doc/stable/user/basics.indexing.html
    #advanced-indexing>`_ as a numpy :py:obj:`~numpy.ndarray`.
    """
    # Attributes displayed:
    _disp_attributes = ('name', 'energy', 'particle', 'periodicity',
                        'harmonic_number', 'beam_current', 'nbunch')
    # excluded attributes
    _excluded_attributes = ('nbunch', )
    # Attributes propagated in copies:
    _std_attributes = ('name', '_energy', '_particle', 'periodicity',
                       '_cell_harmnumber', '_radiation', '_beam_current',
                       '_fillpattern')

    # noinspection PyUnusedLocal
    def __init__(self, *args, iterator: Filter = None, scan: bool = False, **kwargs):
        """
        Lattice(elements, **params)
        Lattice(filter, [filter, ...,] iterator=iter,**params)

        Parameters:
            elements: iterable of Element objects
            iter: function called as :pycode:`iter(params, *args)`. It must
              return an iterable of :py:class:`.Element` objects for building
              the lattice. It must also fill the *params* dictionary providing
              the Lattice attributes.
            params: dictionary of lattice parameters. A custom iterator may
              add, remove or modify parameters. Finally, the remaining
              parameters will be set as Lattice attributes.

        Keyword Arguments:
            name='':        Name of the lattice
            energy:         Energy of the lattice
            periodicity=1:  Number of periods
            particle='relativistic': circulating particle. May be
              'relativistic', 'electron', 'positron', 'proton'
              or a Particle object
            iterator=None:  custom iterator (see below)
            *: all other keywords will be set as attributes of
              the Lattice object
            beam_current:   Total current in the beam, used for collective
              effects

        An iterator ``it`` is called as ``it(params, *args)`` where ``args``
        and ``params`` are the arguments of the ``Lattice`` constructor. It
        must yield the AT ``Elements`` for building the lattice. It must
        also fill its ``params`` dictionary argument, which will be used to
        set the ``Lattice`` attributes.
        An iterator can be:

        - a "generator" which yields elements from scratch.
          Examples: a list, or a file iterator,
        - a "filter" which runs through an input iterator, processes each
          element, possibly adds parameters to the params dictionary
          and yields the processed elements.

        .. Note::

           To reduce the inter-package dependencies, some methods of the
           lattice object are defined in other AT packages, in the module where
           the underlying function is implemented.

        .. Note::

           It is possible to define a filling pattern for the beam using the
           function ``ring.set_fillingpattern()``. The default configuration
           (no arguments) is for single bunch and is the one loaded at lattice
           initialization. See function help for details.
           Changing ``Lattice.harmonic_number`` will reset the filling pattern
           to its default configuration.
           The beam current can be changed with
           ``Lattice.beam_current=current``
           The filling pattern and beam current are used by collective effects
           passmethods.


        Examples:
            Chaining iterators (taken from ``load_mat``):

            .. code-block:: python

               Lattice(ringparam_filter, matfile_generator, filename
                       iterator=params_filter, **params)

            ``matfile_generator(params, filename)``
                opens filename and generates AT elements for each cell of the
                Matlab cell array representing the lattice,

            ``ringparam_filter(params, matfile_generator, *args)``
                runs through ``matfile_generator(params, *args)``, looks for
                RingParam elements, fills params with their information and
                discards them,

            ``params_filter(params, ringparam_filter, *args)``
                runs through ``ringparam_filter(params, *args)``, looks for
                energy and periodicity if not yet defined.
        """
        if iterator is None:
            (arg1,) = args or [[]]  # accept 0 or 1 argument
            if isinstance(arg1, Lattice):
                elems = lattice_filter(kwargs, arg1)
            else:
                elems = params_filter(kwargs, type_filter, arg1)
        else:
            elems = iterator(kwargs, *args)

        super(Lattice, self).__init__(elems)

        # removing excluded attributes
        for attr in self._excluded_attributes:
            kwargs.pop(attr, None)
        # set default values
        kwargs.setdefault("name", "")
        periodicity = kwargs.setdefault("periodicity", 1)
        kwargs.setdefault("particle", kwargs.pop("_particle", Particle()))
        kwargs.setdefault("beam_current", kwargs.pop("_beam_current", 0.0))
        # dummy initialization in case the harmonic number is not there
        kwargs.setdefault("_fillpattern", numpy.ones(1))
        # Remove temporary keywords
        frequency: Optional[float] = kwargs.pop("_frequency", None)
        cell_length: Optional[float] = kwargs.pop("_length", None)
        cell_h = kwargs.pop("cell_harmnumber", kwargs.pop("_cell_harmnumber", math.nan))
        ring_h = kwargs.pop("harmonic_number", cell_h * periodicity)

        energy = kwargs.setdefault("energy", kwargs.pop("_energy", None))
        if energy is None:
            raise AtError("Lattice energy is not defined")

        # set attributes
        self.update(kwargs)

        # Setting the harmonic number is delayed to have self.beta available
        if not (frequency is None or frequency == 0.0):
            rev = self.beta * clight / cell_length
            self._cell_harmnumber = int(round(frequency / rev))
            try:
                fp = kwargs.pop("_fillpattern", numpy.ones(1))
                self.set_fillpattern(bunches=fp)
            except AssertionError:
                self.set_fillpattern()
        elif not math.isnan(ring_h):
            self._cell_harmnumber = ring_h / periodicity

    def __getitem__(self, key):
        try:                                # Integer
            return super(Lattice, self).__getitem__(key.__index__())
        except (AttributeError, TypeError):
            if isinstance(key, slice):      # Slice
                rg = range(*key.indices(len(self)))
            else:                           # Array of integers or boolean
                rg = get_uint32_index(self, key, endpoint=False)
            return Lattice(elem_generator,
                           (super(Lattice, self).__getitem__(i) for i in rg),
                           iterator=self.attrs_filter)

    def __setitem__(self, key, values):
        try:                                # Integer or slice
            super(Lattice, self).__setitem__(key, values)
        except TypeError:                   # Array of integers or boolean
            rg = get_uint32_index(self, key, endpoint=False)
            for i, v in zip(*numpy.broadcast_arrays(rg, values)):
                super(Lattice, self).__setitem__(i, v)

    def __delitem__(self, key):
        try:                                # Integer or slice
            super(Lattice, self).__delitem__(key)
        except TypeError:                   # Array of integers or boolean
            rg = get_uint32_index(self, key, endpoint=False)
            for i in reversed(rg):
                super(Lattice, self).__delitem__(i)

    def __repr__(self):
        ks = ', '.join('{0}={1!r}'.format(k, v) for k, v in self.attrs.items())
        return 'Lattice({0}, {1})'.format(super(Lattice, self).__repr__(), ks)

    def __str__(self):
        ks = ', '.join('{0}={1!r}'.format(k, v) for k, v in self.attrs.items())
        return 'Lattice(<{0} elements>, {1})'.format(len(self), ks)

    def __add__(self, elems):
        """Add elems, an iterable of AT elements, to the lattice"""
        return self.concatenate(elems, copy=True)

    def __iadd__(self, elems):
        self.concatenate(elems, copy=False)
        return self

    def __mul__(self, n):
        return self.repeat(n)

    def _addition_filter(self, elems: Iterable[Element], copy_elements=False):
        cavities = []
        length = 0.0
        params = {}

        for elem in type_filter(params, elems):
            if isinstance(elem, elt.RFCavity):
                cavities.append(elem)
                elem.Energy = self._energy
            elif elem.PassMethod.endswith('RadPass'):
                elem.Energy = self._energy
            elif hasattr(elem, 'Energy'):
                del elem.Energy
            elif hasattr(elem, '_turnhistory'):
                elem.clear_history(self)
            length += getattr(elem, 'Length', 0.0)
            if copy_elements:
                yield elem.deepcopy()
            else:
                yield elem

        if cavities and not hasattr(self, '_cell_harmnumber'):
            cavities.sort(key=lambda el: el.Frequency)
            try:
                self._cell_harmnumber = cavities[0].HarmNumber
            except AttributeError:
                length += self.get_s_pos(len(self))[0]
                rev = self.beta * clight / length
                frequency = cavities[0].Frequency
                self._cell_harmnumber = int(round(frequency / rev))
        self._radiation |= params.pop('_radiation')

    def insert(self, idx: SupportsIndex, elem: Element, copy_elements=False):
        r"""This method allow to insert an AT element in the lattice.

        Parameters:
            idx (SupportsIndex): index at which the lement is inserted
            elem (Element): AT element to be inserted in the lattice
            copy_elements(bool): Default :py:obj:`True`.
                                 If :py:obj:`True` a deep copy of elem
                                 is used.
        """
        # scan the new element to update it
        elist = list(self._addition_filter([elem],  # noqa: F841
                     copy_elements=copy_elements))
        super().insert(idx, elem)

    def extend(self, elems: Iterable[Element], copy_elements=False):
        # noinspection PyUnresolvedReferences
        r"""This method adds all the elements of `elems` to the end of the
        lattice. The behavior is the same as for a :py:obj:`list`

        Equivalent syntaxes:

        >>> ring.extend(elems)
        >>> ring += elems

        Parameters:
            elems (Iterable[Element]): Sequence of AT elements to be
                                      appended to the lattice
            copy_elements(bool): Default :py:obj:`True`.
                                 If :py:obj:`True` deep copies of each
                                 element of elems are used
        """
        if hasattr(self, '_energy'):
            # When unpickling a Lattice, extend is called before the lattice
            # is initialized. So skip this.
            elems = self._addition_filter(elems, copy_elements=copy_elements)
        super().extend(elems)

    def append(self, elem: Element, copy_elements=False):
        # noinspection PyUnresolvedReferences
        r"""This method overwrites the inherited method :py:meth:`list.append()`,
        its behavior is changed, it accepts only AT lattice elements
        :py:obj:`Element` as input argument.

        Equivalent syntaxes:

        >>> ring.append(elem)
        >>> ring += [elem]

        Parameters:
            elem (Element): AT element to be appended to the lattice
            copy_elements(bool): Default :py:obj:`True`.
                                 If :py:obj:`True` a deep copy of elem
                                 is used
        """
        self.extend([elem], copy_elements=copy_elements)

    def repeat(self, n: int, copy_elements: bool = True):
        # noinspection SpellCheckingInspection,PyUnresolvedReferences,PyRedeclaration
        r"""This method allows to repeat the lattice `n` times.
        If `n` does not divide `ring.periodicity`, the new ring
        periodicity is set to 1, otherwise  it is set to
        `ring.periodicity /= n`.

        Equivalent syntaxes:

        >>> newring = ring.repeat(n)
        >>> newring = ring * n

        Parameters:
            n : number of repetitions
            copy_elements:  If :py:obj:`True`, deep copies of the lattice are used for
              the repetition. Otherwise, the original elements are repeated in the
              developed lattice.

        Returns:
            newring (Lattice): the new repeated lattice
        """
        def copy_fun(elem, copy):
            if copy:
                return elem.deepcopy()
            else:
                return elem

        periodicity = self.periodicity
        if n != 0 and periodicity > 1:
            nbp = periodicity / n
            periodicity = int(round(nbp))
            if abs(periodicity - nbp) > _TWO_PI_ERROR:
                warn(AtWarning('Non-integer number of cells: {}/{}. Periodi'
                               'city set to 1'.format(self.periodicity, n)))
                periodicity = 1
        hdict = dict(periodicity=periodicity)
        try:
            hdict.update(harmonic_number=self.cell_harmnumber*n*periodicity)
        except AttributeError:
            pass
        elems = (copy_fun(el, copy_elements) for _ in range(n) for el in self)
        return Lattice(elem_generator, elems, iterator=self.attrs_filter,
                       **hdict)

    def concatenate(self, *lattices: Iterable[Element],
                    copy_elements=False, copy=False):
        # noinspection PyUnresolvedReferences,SpellCheckingInspection,PyRedeclaration
        """Concatenate several `Iterable[Element]` with the lattice

        Equivalent syntaxes:

        >>> newring = ring.concatenate(r1, r2, r3, copy=True)
        >>> newring = ring + r1 + r2 + r3

        >>> ring.concatenate(r1, copy=False)
        >>> ring += r1

        Parameters:
            lattices: :py:obj:`Iterables[Element]` to be concatenanted
                      to the Lattice, several lattices are allowed
                      (see example)
            copy_elements(bool): Default :py:obj:`False`. If :py:obj:`True`
                        deepcopies of the elements of lattices are used
            copy(bool): Default :py:obj:`False`. If :py:obj:`True`
                           the lattice is modified in place.
                           Oterwise a new Lattice object is returned

        Returns:
            lattice(Lattice): concatenated Lattice, if `copy==True` the
                              new lattice object is returned
                              otherwise None
        """
        if copy:
            lattice = Lattice(self)
        else:
            lattice = self
        for lat in lattices:
            lattice.extend(lat, copy_elements=copy_elements)
        return lattice if copy else None

    def reverse(self, copy=False):
        # noinspection PyUnresolvedReferences
        r"""Reverse the order of the lattice and swapt the faces
        of elements. Alignment errors are not swapped

        Parameters:
            copy(bool): Default :py:obj:`False`. If :py:obj:`True`
                           the lattice is modified in place.
                           Oterwise a new Lattice object is returned

        Returns:
            lattice(Lattice): reversed Lattice, if `copy==True` the
                              new lattice object is returned
                              otherwise None
        Example:
            >>> newring = ring.reverse(copy=True)
        """
        elems = (el.swap_faces(copy=True) for el in reversed(self))
        if copy:
            return Lattice(elem_generator, elems, iterator=self.attrs_filter)
        else:
            reversed_list = list(elems)
            self[:] = reversed_list

    def develop(self, copy_elements: bool = True) -> Lattice:
        """Develop a periodical lattice by repeating its elements
        *self.periodicity* times

        Parameters:
            copy_elements:  If :py:obj:`True`, deep copies of the elements are used for
              the repetition. Otherwise, the original elements are repeated in the
              developed lattice.

        Returns:
            newlattice: The developed lattice
        """
        return self.repeat(self.periodicity, copy_elements=copy_elements)

    @property
    def attrs(self) -> dict:
        """Dictionary of lattice attributes"""
        def extattr(d):
            for k in self._disp_attributes:
                d.pop(k, None)
                yield k, getattr(self, k, None)

        vrs = vars(self).copy()
        # Standard attributes
        res = {k: v for k, v in extattr(vrs) if v is not None}
        # Custom attributes
        res.update((k, v) for k, v in vrs.items() if not k.startswith('_'))
        return res

    def rotate(self, n: int) -> Lattice:
        """Return a new lattice rotated left by n elements"""
        if len(self) == 0:
            return self.copy()
        n = n % len(self)      # works for n<0
        return self[n:] + self[:n]

    def update(self, *args, **kwargs) -> None:
        """
        update(**kwargs)
        update(mapping, **kwargs)
        update(iterable, **kwargs)

        Update the lattice attributes with the given values
        """
        attrs = dict(*args, **kwargs)
        for (key, value) in attrs.items():
            setattr(self, key, value)

    def copy(self) -> Lattice:
        """Returns a shallow copy of the lattice"""
        return copy.copy(self)

    def deepcopy(self) -> Lattice:
        """Returns a deep copy of the lattice"""
        return copy.deepcopy(self)

    def slice_elements(self, refpts: Refpts, slices: int = 1) -> Lattice:
        """Create a new lattice by slicing the elements at refpts

        Parameters:
            refpts:     Element selector
            slices:     Number of slices in the specified range. Ignored if
              size is specified. Default: no slicing

        Returns:
            newring:    New Lattice object
        """
        def slice_generator(_):
            check = get_bool_index(self, refpts)
            for el, ok in zip(self, check):
                if ok and (slices > 1):
                    frac = numpy.ones(slices) / slices
                    for elem in el.divide(frac):
                        yield elem
                else:
                    yield el

        return Lattice(slice_generator, iterator=self.attrs_filter)

    def slice(self, size: Optional[float] = None, slices: Optional[int] = 1) \
            -> Lattice:
        """Create a new lattice by slicing the range of interest into small
        elements

        Keyword arguments:
            size=None:      Length of a slice. Default: computed from the
              range and number of points: ``size = (s_max-s_min)/slices``.
            slices=1:       Number of slices in the specified range. Ignored if
              size is specified. Default: no slicing

        Returns:
            newring:    New Lattice object
        """
        def slice_generator(_, ibeg, iend):
            yield from self[:ibeg]
            for elem in self[ibeg:iend]:
                nslices = int(math.ceil(elem.Length / size))
                if nslices > 1:
                    frac = numpy.ones(nslices) / nslices
                    for el in elem.divide(frac):
                        yield el
                else:
                    yield elem
            yield from self[iend:]

        s_range = self.s_range
        if size is None:
            smin, smax = s_range
            size = (smax - smin) / slices
        i_range = self.i_range
        return Lattice(slice_generator, i_range[0], i_range[-1],
                       iterator=self.attrs_filter, s_range=s_range)

    def attrs_filter(self, params: dict, elem_filter: Filter, *args) \
            -> Iterable[Element]:
        """Filter function which duplicates the lattice attributes

        Parameters:
            params:         Dictionary of Lattice attributes
            elem_filter:    Element filter
            *args:          Arguments provided to *elem_filter*
        """
        for key in self._std_attributes:
            try:
                params.setdefault(key, getattr(self, key))
            except AttributeError:
                pass
        return elem_filter(params, *args)

    @property
    def s_range(self) -> Union[None, tuple[float, float]]:
        """Range of interest: (s_min, s_max). :py:obj:`None` means
          the full cell."""
        try:
            return self._s_range
        except AttributeError:
            self.s_range = None
            return self._s_range

    # noinspection PyAttributeOutsideInit
    @s_range.setter
    def s_range(self, value: Optional[tuple[float, float]]):
        spos = self.get_s_pos(range(len(self) + 1))
        if value is None:
            value = (0.0, spos[-1])
        ok = numpy.flatnonzero(
            numpy.logical_and(spos > value[0], spos < value[1]))
        if len(ok) > 0:
            i1 = max(ok[0] - 1, 0)
            i2 = min(ok[-1] + 1, len(self))
            self._i_range = range(i1, i2 + 1)
        else:
            self._i_range = range(0, 0)
        self._s_range = value

    @property
    def i_range(self) -> Uint32Refpts:
        """Range of elements inside the range of interest"""
        try:
            i_range = self._i_range
        except AttributeError:
            self.s_range = None
            i_range = self._i_range
        return get_uint32_index(self, i_range)

    @property
    def energy(self) -> float:
        """Lattice energy"""
        return self._energy

    @energy.setter
    def energy(self, energy: float):
        # Set the Energy attribute of radiating elements
        for elem in self:
            if (isinstance(elem, (elt.RFCavity, elt.Wiggler))
                    or elem.PassMethod.endswith('RadPass')):
                elem.Energy = energy
        # Set the energy attribute of the Lattice
        # Use a numpy scalar to allow division by zero
        # self._energy = numpy.array(energy, dtype=float)
        self._energy = energy

    @property
    def cell_length(self) -> float:
        """Cell length (1 cell) [m]

        See Also:

            :py:meth:`circumference`.
        """
        return self.get_s_pos(len(self))[0]

    @property
    def cell_revolution_frequency(self) -> float:
        """Revolution frequency for 1 cell [Hz]

        See Also:

            :py:meth:`revolution_frequency`.
        """
        beta = self.beta
        return beta * clight / self.cell_length

    @property
    def cell_harmnumber(self) -> float:
        """Harmonic number per cell

        See Also:

            :py:meth:`harmonic_number`.
        """
        try:
            return self._cell_harmnumber
        except AttributeError:
            raise AttributeError('harmonic_number undefined: '
                                 'No cavity found in the lattice') from None

    @property
    def circumference(self) -> float:
        """Ring circumference (full ring) [m]

        See Also:

            :py:meth:`cell_length`.
        """
        return self.periodicity * self.get_s_pos(len(self))[0]

    @property
    def revolution_frequency(self) -> float:
        """Revolution frequency (full ring) [Hz]

        See Also:

            :py:meth:`cell_revolution_frequency`.
        """
        beta = self.beta
        return beta * clight / self.circumference

    @property
    def particle(self) -> Particle:
        """Circulating particle"""
        return self._particle

    @particle.setter
    def particle(self, particle: Union[str, Particle]):
        if isinstance(particle, Particle):
            self._particle = particle
        else:
            self._particle = Particle(particle)

    def set_wake_turnhistory(self):
        """Function to reset the shape of the turn history
        in collective elements based on the number of slices,
        turns and bunches
        """
        for e in self:
            if e.is_collective:
                e.clear_history(ring=self)

    def set_fillpattern(self, bunches: Union[int, numpy.ndarray] = 1):
        """Function to generate the filling pattern lof the ring.
        The filling pattern is computed as:

        ``bunches/numpy.sum(bunches)``

        This function also generates the bunch spatial distribution
        accessible with ``Lattice.bunch_spos``

        Keyword Arguments:
           bunches:  integer or array of positive double or bool to define
                     the bunch distribution.
                     For scalar input, equidistant bunches are assumed.
                     ``ring.harmonic_number`` has to be a multiple of
                     ``bunches``.
                     For array input the condition
                     ``len(bunches)==ring.harmonic_number`` is required.
                     (default=1, single bunch configuration).
        """
        if isinstance(bunches, int):
            if self.harmonic_number % bunches == 0:
                fp = numpy.zeros(self.harmonic_number)
                fp[::int(self.harmonic_number/bunches)] = 1
            else:
                raise AtError('Harmonic number has to be a '
                              'multiple of the scalar input '
                              'bunches')
        elif numpy.isscalar(bunches):
            raise AtError('Scalar input for bunches must be '
                          'an integer')
        else:
            bunches = bunches.astype(dtype=float, casting='safe',
                                     copy=False)
            assert len(bunches) == self.harmonic_number, \
                'bunches array input has to be of shape ({0},)' \
                .format(self.harmonic_number)
            assert numpy.all(bunches >= 0.0), \
                'bunches array can contain only positive numbers'
            fp = bunches
        self._fillpattern = fp/numpy.sum(fp)
        self.set_wake_turnhistory()

    @property
    def fillpattern(self) -> numpy.ndarray:
        """Filling pattern describing the bunch relative
        amplitudes such that ``sum(fillpattern)=1``
        """
        return self._fillpattern

    @fillpattern.setter
    def fillpattern(self, value):
        """Filling pattern describing the bunch relative
        amplitudes such that ``sum(fillpattern)=1``.
        Calls the function ``Lattice.set_fillpattern``.
        """
        self.set_fillpattern(value)

    @property
    def bunch_list(self) -> numpy.ndarray:
        """Indices of filled bunches"""
        return numpy.flatnonzero(self._fillpattern)

    def get_beam_current(self):
        return self._beam_current

    def set_beam_current(self, value, clear_history=True):
        self._beam_current = value
        if clear_history:
            self.set_wake_turnhistory()

    beam_current = property(get_beam_current, set_beam_current)

    @property
    def bunch_currents(self) -> numpy.ndarray:
        """Bunch currents [A]"""
        return self.beam_current * \
            self._fillpattern[self._fillpattern > 0]

    @property
    def bunch_spos(self) -> numpy.ndarray:
        """Bunch position around the ring [m]"""
        try:
            circ = self.beta * clight * \
                self.harmonic_number/self.rf_frequency
        except AtError:
            circ = self.circumference
        bs = circ/len(self._fillpattern)
        allpos = bs*numpy.arange(len(self._fillpattern))
        return allpos[self._fillpattern > 0]

    @property
    def nbunch(self) -> int:
        """Number of bunches"""
        return numpy.count_nonzero(self._fillpattern)

    @property
    def harmonic_number(self) -> int:
        """Ring harmonic number (full ring)

        See Also:

            :py:meth:`cell_harmnumber`.
        """
        try:
            return int(self.periodicity * self._cell_harmnumber)
        except AttributeError:
            raise AttributeError('harmonic_number undefined: '
                                 'No cavity found in the lattice') from None

    @harmonic_number.setter
    def harmonic_number(self, value):
        cell_h = float(value) / self.periodicity
        # check on ring
        if value-round(value) != 0:
            raise AtError('harmonic number ({}) must be integer'.format(value))
        # check on cell
        # if cell_h-round(cell_h) != 0:
        #     raise AtError('harmonic number ({}) must be a multiple of {}'
        #                   .format(value, int(self.periodicity)))
        self._cell_harmnumber = cell_h
        if len(self._fillpattern) != value:
            warn(AtWarning('Harmonic number changed, resetting fillpattern to '
                           'default (single bunch)'))
            self.set_fillpattern()

    @property
    def gamma(self) -> float:
        r"""Relativistic :math:`\gamma` of the particles"""
        rest_energy = self.particle.rest_energy
        if rest_energy == 0.0:
            rest_energy = e_mass
        return float(self.energy / rest_energy)

    @property
    def beta(self) -> float:
        r"""Relativistic :math:`\beta` of the particles"""
        rest_energy = self.particle.rest_energy
        if rest_energy == 0.0:
            return 1.0
        else:
            gamma = float(self.energy / rest_energy)
            return math.sqrt(1.0 - 1.0/gamma/gamma)

    # noinspection PyPep8Naming
    @property
    def BRho(self) -> float:
        """Magnetic rigidity [T.m]"""
        rest_energy = self.particle.rest_energy
        if rest_energy == 0.0:
            rest_energy = e_mass
        return math.sqrt(self.energy**2 - rest_energy**2)/clight

    @property
    def is_6d(self) -> bool:
        """True if at least one element modifies the beam momentum

        See Also:

            :py:meth:`enable_6d`, :py:meth:`disable_6d`.
        """
        try:
            return self._radiation
        except AttributeError:
            radiate = False
            for elem in self:
                if elem.longt_motion:
                    radiate = True
                    break
            # noinspection PyAttributeOutsideInit
            self._radiation = radiate
            return radiate

    @property
    def is_collective(self) -> bool:
        """:py:obj:`True` if any element involves collective effects"""
        for elem in self:
            if elem.is_collective:
                return True
        return False

    @property
    def has_cavity(self) -> bool:
        """:py:obj:`True` if the lattice contains an active
        :py:class:`RFCavity`
        """
        for elem in self:
            if elem.PassMethod.endswith('CavityPass'):
                return True
        return False

    # noinspection PyShadowingNames
    def modify_elements(self, elem_modify: Callable,
                        copy: Optional[bool] = True, **kwargs):
        """Modify selected elements, in-place or in a lattice copy

        Parameters:
            elem_modify : element selection function.
              If ``elem_modify(elem)`` returns :py:obj:`None`, the element is
              unchanged. Otherwise, ``elem_modify(elem)`` must return a
              dictionary of attribute name and values, to be set to elem.
            copy:         If True, return a shallow copy of the lattice.
              Only the modified elements are copied.
              If False, the modification is done in-place

        Returns:
            newring:    New lattice if copy is :py:obj:`True`, :py:obj:`None`
              if copy is :py:obj:`False`

        Keyword Arguments:
        """

        def lattice_modify(**kws):
            """Modifies the Lattice with elements modified by elem_modify"""
            radiate = False
            for elem in self:
                attrs = elem_modify(elem)
                if attrs is not None:
                    elem.update(attrs)
                if elem.longt_motion:
                    radiate = True
            self._radiation = radiate
            self.update(kws)

        def lattice_copy(params):
            """Custom iterator for the creation of a new lattice"""
            radiate = False
            for elem in self:
                attrs = elem_modify(elem)
                if attrs is not None:
                    elem = elem.copy()
                    elem.update(attrs)
                if elem.longt_motion:
                    radiate = True
                yield elem
            params['_radiation'] = radiate

        if copy:
            return Lattice(lattice_copy, iterator=self.attrs_filter, **kwargs)
        else:
            lattice_modify(**kwargs)

    def _set_6d(self, enable: bool, *args, **kwargs):
        """Set the lattice radiation state"""

        def lattice_modify():
            """Modifies the Lattice in place"""
            radiate = False
            for elem in self:
                new_pass = getpass(elem)
                if new_pass:
                    elem.set_longt_motion(enable, new_pass=new_pass, **vargs)
                if elem.longt_motion:
                    radiate = True
            self._radiation = radiate
            self.update(kwargs)

        def lattice_copy(params):
            """Custom iterator for the creation of a new lattice"""
            radiate = False
            for elem in self:
                new_pass = getpass(elem)
                if new_pass:
                    elem = elem.set_longt_motion(enable, new_pass=new_pass,
                                                 copy=True, **vargs)
                if elem.longt_motion:
                    radiate = True
                yield elem
            params['_radiation'] = radiate

        cp = kwargs.pop('copy', False)
        if len(args) > 0:
            def getpass(elem):
                return 'auto' if isinstance(elem, args) else None

            if not all(issubclass(cl, elt.LongtMotion) for cl in args):
                raise TypeError("All arguments must be subclasses of"
                                " 'LongtMotion'")
            if len(kwargs) > 0:
                raise AtError('No keyword is allowed in this mode')
        else:
            def getpass(elem):
                for eltype, psm in pass_table:
                    if isinstance(elem, eltype):
                        return psm
                return None

            def passm(key, eltype, def_pass):
                if allset:
                    def_pass = all_pass
                return eltype, kwargs.pop(key, def_pass)

            # Look for global defaults
            try:
                all_pass = kwargs.pop('all_pass')
            except KeyError:
                allset = False
            else:
                allset = True
            # Build table of PassMethods
            pass_table = [passm(*defs) for defs in _DEFAULT_PASS[enable]]

        vargs = dict(energy=self.energy) if enable else {}
        if cp:
            return Lattice(lattice_copy, iterator=self.attrs_filter, **kwargs)
        else:
            lattice_modify()

    # noinspection PyShadowingNames,PyIncorrectDocstring
    def enable_6d(self, *args, **kwargs) -> Optional[Lattice]:
        # noinspection PyUnresolvedReferences
        r"""
        enable_6d(elem_class[, elem_class]..., copy=False)
        enable_6d(cavity_pass='auto'[, dipole_pass='auto']..., copy=False)

        Turn longitudinal motion on. By default, turn on
        radiation in dipoles and quadrupoles, turn on RF cavities, activates
        collective effects and other elements acting on momentum.

        Modify the lattice in-place or creates a new lattice, depending on the
        ``copy`` keyword argument.

        **Syntax using positional arguments:**

        Parameters:
            elem_class:                 :py:class:`.LongtMotion` subclass.
              Longitudinal motion is turned on for elements which are
              instances of any given ``elem_class``.

              In adition to single element classes, a few grouping classes are
              available:

              * :py:class:`.LongtMotion`: all elements possibly acting on
                momentum,
              * :py:class:`.Radiative`: default radiative elements:
                :py:class:`.Dipole`, :py:class:`.Quadrupole`,
                :py:class:`.Wiggler`,
              * :py:class:`.Collective`: all elements dealing with collective
                effects.

              The default PassMethod conversion is used, as with the ``'auto'``
              keyword value..

              **No keyword except** ``copy`` **is allowed in this syntax.**

        **Syntax using keyword arguments:**

        Keyword arguments:
            all_pass:                   PassMethod overloading the default
              values for all elements (:py:obj:`None` or 'auto')
            cavity_pass='auto':         PassMethod set on cavities
            dipole_pass='auto':         PassMethod set on dipoles
            quadrupole_pass='auto':     PassMethod set on quadrupoles
            wiggler_pass='auto':        PassMethod set on wigglers
            sextupole_pass=None:        PassMethod set on sextupoles
            octupole_pass=None:         PassMethod set on octupoles
            multipole_pass=None  :      PassMethod set on higher order
              multipoles
            collective_pass='auto':     PassMethod set on collective effect
              elements (:py:class:`.WakeElement`,...)
            diffusion_pass='auto':      PassMethod set on
              :py:class:`.QuantumDiffusion`
            copy=False: If :py:obj:`False`, the modification is done in-place,
              If ``True``, return a shallow copy of the lattice. Only the
              radiating elements are copied with PassMethod modified.

              .. Caution::

                 a shallow copy means that all non-modified elements are shared
                 with the original lattice. Any further modification will
                 affect both lattices.

        For PassMethod names, the convention is:

        * :py:obj:`None`:        No change
        * 'auto':          Use the default conversion (replace \*Pass by
          \*RadPass)
        * anything else:   set as the new PassMethod

        Examples:
            >>> ring.enable_6d()

            Modify `ring` in-place, turn cavities ON, turn synchrotron
            radiation ON in Dipoles and Quadrupoles, turn collective effects
            ON.

            >>> ring.enable_6d(at.RFCavity, at.Radiative)

            Modify `ring` in-place, turn cavities ON, turn synchrotron
            radiation ON in dipoles and quadupoles.

            >>> newring = ring.enable_6d(at.Collective, copy=True)

            Returns a new lattice with collective effects turned ON and nothing
            else changed

            >>> newring = ring.enable_6d(all_pass=None, collective_pass='auto',
            ... copy=True)

            Same as the previous example, using the keyword syntax.

        See Also:

            :py:meth:`disable_6d`, :py:attr:`is_6d`.
        """
        return self._set_6d(True, *args, **kwargs)

    # noinspection PyShadowingNames,PyIncorrectDocstring
    def disable_6d(self, *args, **kwargs) -> Optional[Lattice]:
        # noinspection PyUnresolvedReferences
        r"""
        disable_6d(elem_class[, elem_class]... , copy=False)
        disable_6d(cavity_pass='auto'[, dipole_pass='auto']..., copy=False)

        Turn longitudinal motion off. By default, remove all longitudinal
        motion.

        Modify the lattice in-place or creates a new lattice, depending on the
        ``copy`` keyword argument.

        **Syntax using positional arguments:**

        Parameters:
            elem_class:                 :py:class:`.LongtMotion` subclass.
              Longitudinal motion is turned off for elements which are
              instances of any given ``elem_class``.

              In adition to single element classes, a few grouping classes are
              available:

              * :py:class:`.LongtMotion`: all elements possibly acting on
                momentum,
              * :py:class:`.Radiative`: default radiative elements:
                :py:class:`.Dipole`, :py:class:`.Quadrupole`,
                :py:class:`.Wiggler`,
              * :py:class:`.Collective`: all elements dealing with collective
                effects.

              The default PassMethod conversion is used, as with the ``'auto'``
              keyword value.

              **No keyword except** ``copy`` **is allowed in this syntax.**

        **Syntax using keyword arguments:**

        Keyword arguments:
            all_pass:                   PassMethod overloading the default
              values for all elements (:py:obj:`None` or 'auto')
            cavity_pass='auto':         PassMethod set on cavities
            dipole_pass='auto':         PassMethod set on dipoles
            quadrupole_pass=auto:       PassMethod set on quadrupoles
            wiggler_pass='auto':        PassMethod set on wigglers
            sextupole_pass='auto':      PassMethod set on sextupoles
            octupole_pass='auto':       PassMethod set on octupoles
            multipole_pass='auto':      PassMethod set on higher order
              multipoles
            collective_pass='auto':     PassMethod set on collective effect
              elements (:py:class:`.WakeElement`,...)
            diffusion_pass='auto':      PassMethod set on
              :py:class:`.QuantumDiffusion`
            copy=False: If :py:obj:`False`, the modification is done in-place,
              If ``True``, return a shallow copy of the lattice. Only the
              radiating elements are copied with PassMethod modified.

              .. Caution::

                 a shallow copy means that all non-modified elements are shared
                 with the original lattice. Any further modification will
                 affect both lattices.

        For PassMethod names, the convention is:

        * :py:obj:`None`:        no change
        * 'auto':          Use the default conversion (replace \*RadPass by
          \*Pass)
        * anything else:   set as the new PassMethod

        Examples:
            >>> ring.disable_6d()

            Modify `ring` in-place, turn OFF everything affecting the
            longitudinal momentum.

            >>> ring.disable_6d(at.RFCavity)

            Turn cavities OFF (nothing else modified).

            >>> ring.disable_6d(all_pass=None, cavity_pass='auto')

            Same as the previous example, but using the keyword syntax.

            >>> newring = ring.disable_6d(cavity_pass=None, copy=True)

            Return a new Lattice (shallow copy of `ring`) with everything
            turned OFF except RF cavities.

            >>> newring = ring.disable_6d(all_pass=None,
            ... sextupole_pass='DriftPass', copy=True)

            Return a new Lattice (shallow copy of `ring`) with sextupoles
            turned into Drifts (turned off) and everything else unchangedd.

        See Also:

            :py:meth:`enable_6d`, :py:attr:`is_6d`.
        """
        return self._set_6d(False, *args, **kwargs)

    def sbreak(self, break_s, break_elems=None, **kwargs):
        """Insert elements at selected locations in the lattice

        Parameters:
            break_s:        location or array of locations of breakpoints
            break_elems:    elements to be inserted at breakpoints (array of
                            elements as long as break_s or single element
                            duplicated as necessary). Default: Marker('sbreak')
        Returns:
            newring:    A new lattice with new elements inserted at breakpoints
        """

        def sbreak_iterator(_, itmk):
            """Iterate over elements and breaks where necessary"""

            def next_mk():
                """Extract the next element to insert"""
                try:
                    return next(itmk)
                except StopIteration:
                    return sys.float_info.max, None

            s_end = 0.0
            # get the 1st insertion
            smk, mk = next_mk()
            # skip all insertions at negative break_s, if any
            while smk < s_end:
                smk, mk = next_mk()

            for elem in self:
                s_end += elem.Length
                # loop over all insertions within the element
                while smk < s_end:
                    frac = (s_end - smk) / elem.Length
                    if frac < 1.0:  # breakpoint is within the element
                        el0, elem = elem.divide([1.0 - frac, frac])
                        yield el0
                    yield mk
                    smk, mk = next_mk()
                yield elem

        # set default insertion
        if break_elems is None:
            break_elems = elt.Marker('sbreak')
        break_elems = numpy.reshape(break_elems, -1)
        # Check element lengths
        if not all(e.Length == 0 for e in break_elems):
            warn(AtWarning(
                 "Inserting elements with length!=0 may change the lattice"))
        # broadcast break_s and break_elems to arrays of same size
        # and create an iterator over the elements to be inserted
        iter_mk = zip(*numpy.broadcast_arrays(break_s, break_elems))

        return Lattice(sbreak_iterator, iter_mk,
                       iterator=self.attrs_filter, **kwargs)

    def reduce(self, **kwargs) -> Lattice:
        """Removes all elements with an ``IdentityPass`` PassMethod and merges
        compatible consecutive elements.

        The "reduced" lattice has the same optics as the original one, but
        fewer elements. Compatible elements must have the same ``PolynomA``,
        ``PolynomB`` and bending radius, so that the optics is preserved. But
        magnet misalignments are not preserved, so this method should be
        applied to lattices without errors.

        Keyword Args:
            keep (Refpts):      Keep the selected elements, even with
                ``IdentityPass`` PassMethod. Default: keep :py:class:`.Monitor`
                and :py:class:`.RFCavity` elements
        """
        def reduce_filter(_, itelem):
            try:
                elem = next(itelem).copy()
            except StopIteration:
                return
            for nxt in itelem:
                try:
                    elem.merge(nxt)
                except TypeError:
                    yield elem
                    elem = nxt.copy()
            yield elem

        kp = get_bool_index(self, lambda el: el.PassMethod != 'IdentityPass')
        try:
            keep = get_bool_index(self, kwargs.pop('keep'))
        except KeyError:
            keep = get_bool_index(self, checktype((elt.Monitor, elt.RFCavity)))

        return Lattice(reduce_filter, self.select(kp | keep),
                       iterator=self.attrs_filter, **kwargs)

    def replace(self, refpts: Refpts, **kwargs) -> Lattice:
        """Return a shallow copy of the lattice replacing the selected
        elements by a deep copy

        Parameters:
            refpts: element selector
        """
        check = get_bool_index(self, refpts)
        elems = (el.deepcopy() if ok else el for el, ok in zip(self, check))
        return Lattice(elem_generator, elems,
                       iterator=self.attrs_filter, **kwargs)

    # Obsolete methods kept for compatibility
    def radiation_on(self, *args, **kwargs) -> Optional[Lattice]:
        """Obsolete. Turn longitudinal motion on

        The function name is misleading, since the function deals with
        longitudinal motion in general.

        For this reason **the method is obsolete** and replaced by
        :py:meth:`.enable_6d`

        See Also:
            :py:meth:`.enable_6d`
        """
        kwargs.update(
            zip(('cavity_pass', 'dipole_pass', 'quadrupole_pass'), args))
        return self._set_6d(True, **kwargs)

    def radiation_off(self, *args, **kwargs) -> Optional[Lattice]:
        """Obsolete. Turn longitudinal motion off

        The function name is misleading, since the function deals with
        longitudinal motion in general.

        For this reason **the method is obsolete** and replaced by
        :py:meth:`.disable_6d`

        See Also:
            :py:meth:`disable_6d`
        """
        kwargs.update(
            zip(('cavity_pass', 'dipole_pass', 'quadrupole_pass'), args))
        return self._set_6d(False, **kwargs)

    @property
    def radiation(self) -> bool:
        """Obsolete (see :py:attr:`is_6d` instead)

        :meta private:
        """
        try:
            return self._radiation
        except AttributeError:
            radiate = False
            for elem in self:
                if elem.longt_motion:
                    radiate = True
                    break
            # noinspection PyAttributeOutsideInit
            self._radiation = radiate
            return radiate


def lattice_filter(params, lattice):
    """Copy lattice parameters and run through all lattice elements

    Parameters:
        params:     Dictionary of Lattice attributes
        lattice:    Input ``Lattice``

    Yields:
        lattice ``Elements``
    """
    return lattice.attrs_filter(params, elem_generator, lattice)


# noinspection PyUnusedLocal
def elem_generator(params, elems: Iterable[Element]) -> Iterable[Element]:
    """Run through all elements without any check

    Parameters:
        params:     Dictionary of Lattice attributes
        elems:      Iterable of lattice ``Elements``

    Yields:
        lattice ``Elements``
    """
    return elems


no_filter: Filter = elem_generator  # provided for backward compatibility


def type_filter(params, elems: Iterable[Element]) -> Generator[Element, None, None]:
    """Run through all elements and check element validity.
    Analyse elements for radiation state

    Parameters:
        params:     Dictionary of Lattice attributes
        elems:      Iterable of lattice ``Elements``

    Yields:
        lattice ``Elements``
    """
    radiate = False
    for idx, elem in enumerate(elems):
        if isinstance(elem, Element):
            if elem.longt_motion:
                radiate = True
            yield elem
        else:
            warn(AtWarning('item {0} ({1}) is not an AT element: '
                           'ignored'.format(idx, elem)))
    params['_radiation'] = radiate


def params_filter(params, elem_filter: Filter, *args) -> Generator[Element, None, None]:
    """Run through all elements, looking for energy and periodicity.
    Remove the Energy attribute of non-radiating elements

    Parameters:
        params:         Dictionary of Lattice attributes
        elem_filter:    Next ``Elements`` filter
        args:           Arguments forwarded to **elem_filter**

    Yields:
        lattice ``Elements``

    The following keys in ``params`` are set:

    * ``_length``
    * ``periodicity``
    * ``energy`` (optional)
    * ``_frequency`` (optional)
    * ``harmonic_number`` (optional)

    energy is taken from:
        1) The params dictionary
        2) Cavity elements
        3) Any other element

    periodicity is taken from:
        1) The params dictionary
        2) Sum of the bending angles of magnets
    """
    el_energies = []
    thetas = []
    cavities = []
    cell_length = 0

    for elem in elem_filter(params, *args):
        if isinstance(elem, elt.RFCavity):
            cavities.append(elem)
        elif hasattr(elem, 'Energy'):
            el_energies.append(elem.Energy)
            del elem.Energy
        if isinstance(elem, elt.Dipole):
            thetas.append(elem.BendingAngle)
        cell_length += getattr(elem, 'Length', 0.0)
        yield elem

    params['_length'] = cell_length
    cav_energies = [el.Energy for el in cavities if hasattr(el, 'Energy')]
    if cavities:
        cavities.sort(key=lambda el: el.Frequency)
        c0 = cavities[0]
        params.setdefault('harmonic_number',
                          getattr(c0, 'HarmNumber', math.nan))
        # params['_harmnumber'] = getattr(c0, 'HarmNumber', math.nan)
        params['_frequency'] = getattr(c0, 'Frequency', None)

    if 'energy' not in params:
        energies = cav_energies or el_energies
        if energies:
            # Guess energy from the Energy attribute of the elements
            energy = params.setdefault('energy',  max(energies))
            if min(energies) < max(energies):
                warn(AtWarning('Inconsistent energy values, '
                               '"energy" set to {0}'.format(energy)))

    if 'periodicity' not in params:
        # Guess periodicity from the bending angles of the lattice
        try:
            nbp = 2.0 * numpy.pi / sum(thetas)
        except ZeroDivisionError:
            periodicity = 1
            warn(AtWarning('No bending in the cell, set "Periodicity" to 1'))
        else:
            periodicity = int(round(nbp))
            if abs(periodicity - nbp) > _TWO_PI_ERROR:
                warn(AtWarning('Non-integer number of cells: '
                               '{0} -> {1}'.format(nbp, periodicity)))
        params['periodicity'] = periodicity


Lattice.get_uint32_index = get_uint32_index
Lattice.get_bool_index = get_bool_index
Lattice.refcount = _refcount
Lattice.set_shift = set_shift
Lattice.set_tilt = set_tilt
Lattice.get_elements = get_elements
Lattice.get_s_pos = get_s_pos
Lattice.select = refpts_iterator
Lattice.get_value_refpts = get_value_refpts
Lattice.set_value_refpts = set_value_refpts
Lattice.get_geometry = get_geometry



"""
Import/export AT lattice from/to different formats:
- .mat files
- .m files
"""
from .allfiles import *
from .matfile import *
from .reprfile import *
from .tracy import *
from .elegant import *
from .json import *
